from OpenGL import GL, constant
import OpenGL.GL.ARB.gpu_shader_int64 as gsi64
import numpy as np
import glm
import ctypes

dtype_uint8 = np.array([], dtype=np.uint8).dtype
dtype_int8 = np.array([], dtype=np.int8).dtype
dtype_uint16 = np.array([], dtype=np.uint16).dtype
dtype_int16 = np.array([], dtype=np.int16).dtype
dtype_uint32 = np.array([], dtype=np.uint32).dtype
dtype_uint64 = np.array([], dtype=np.uint64).dtype
dtype_int32 = np.array([], dtype=np.int32).dtype
dtype_int64 = np.array([], dtype=np.int64).dtype
dtype_float16 = np.array([], dtype=np.float16).dtype
dtype_float32 = np.array([], dtype=np.float32).dtype
dtype_float64 = np.array([], dtype=np.float64).dtype

class GLInfo:
	primary_types = \
	[
		int, float, bool,

		np.int8, np.uint8,
		np.int16, np.uint16,
		np.int32, np.uint32,
		np.int64, np.uint64,
		np.float16, np.float32, np.float64,

		dtype_int8, dtype_uint8,
		dtype_int16, dtype_uint16,
		dtype_int32, dtype_uint32,
		dtype_int64, dtype_uint64,
		dtype_float16, dtype_float32, dtype_float64,

		ctypes.c_int8, ctypes.c_uint8,
		ctypes.c_int16, ctypes.c_uint16,
		ctypes.c_int32, ctypes.c_uint32,
		ctypes.c_int64, ctypes.c_uint64,
		ctypes.c_float, ctypes.c_double
	]

	dtype_map = \
	{
		GL.GL_BYTE: np.int8,
		GL.GL_UNSIGNED_BYTE: np.uint8,
		GL.GL_SHORT: np.int16,
		GL.GL_UNSIGNED_SHORT: np.uint16,
		GL.GL_INT: np.int32,
		GL.GL_UNSIGNED_INT: np.uint32,
		gsi64.GL_UNSIGNED_INT64_ARB: np.uint64,
		GL.GL_HALF_FLOAT: np.float16,
		GL.GL_FLOAT: np.float32,
		GL.GL_DOUBLE: np.float64,
	}

	dtype_inverse_map = \
	{
		dtype_int8:GL.GL_BYTE, np.int8:GL.GL_BYTE, ctypes.c_int8:GL.GL_BYTE, bool:GL.GL_BYTE,
		dtype_uint8:GL.GL_UNSIGNED_BYTE, np.uint8:GL.GL_UNSIGNED_BYTE, ctypes.c_uint8:GL.GL_UNSIGNED_BYTE,
		dtype_int16:GL.GL_SHORT, np.int16:GL.GL_SHORT, ctypes.c_int16:GL.GL_SHORT,
		dtype_uint16:GL.GL_UNSIGNED_SHORT, np.uint16:GL.GL_UNSIGNED_SHORT, ctypes.c_uint16:GL.GL_UNSIGNED_SHORT,
		dtype_int32:GL.GL_INT, np.int32:GL.GL_INT, ctypes.c_int32:GL.GL_INT, int:GL.GL_INT,
		dtype_uint32:GL.GL_UNSIGNED_INT, np.uint32:GL.GL_UNSIGNED_INT, ctypes.c_uint32:GL.GL_UNSIGNED_INT,
		dtype_uint64:gsi64.GL_UNSIGNED_INT64_ARB, np.uint64:gsi64.GL_UNSIGNED_INT64_ARB, ctypes.c_uint64:gsi64.GL_UNSIGNED_INT64_ARB,
		dtype_float16:GL.GL_HALF_FLOAT, np.float16:GL.GL_HALF_FLOAT,
		dtype_float32:GL.GL_FLOAT, np.float32:GL.GL_FLOAT, ctypes.c_float:GL.GL_FLOAT, float:GL.GL_FLOAT,
		dtype_float64:GL.GL_DOUBLE, np.float64:GL.GL_DOUBLE, ctypes.c_double:GL.GL_DOUBLE,
		glm.vec2: GL.GL_FLOAT, glm.vec3: GL.GL_FLOAT, glm.vec4: GL.GL_FLOAT,
		glm.bvec2: GL.GL_BYTE, glm.bvec3: GL.GL_BYTE, glm.bvec4: GL.GL_BYTE,
		glm.ivec2: GL.GL_INT, glm.ivec3: GL.GL_INT, glm.ivec4: GL.GL_INT,
		glm.uvec2: GL.GL_UNSIGNED_INT, glm.uvec3: GL.GL_UNSIGNED_INT, glm.uvec4: GL.GL_UNSIGNED_INT,
		glm.dvec2: GL.GL_DOUBLE, glm.dvec3: GL.GL_DOUBLE, glm.dvec4: GL.GL_DOUBLE,
		glm.mat2x2: GL.GL_FLOAT, glm.mat2x3: GL.GL_FLOAT, glm.mat2x4: GL.GL_FLOAT,
		glm.mat3x2: GL.GL_FLOAT, glm.mat3x3: GL.GL_FLOAT, glm.mat3x4: GL.GL_FLOAT,
		glm.mat4x2: GL.GL_FLOAT, glm.mat4x3: GL.GL_FLOAT, glm.mat4x4: GL.GL_FLOAT,
		glm.dmat2x2: GL.GL_DOUBLE, glm.dmat2x3: GL.GL_DOUBLE, glm.dmat2x4: GL.GL_DOUBLE,
		glm.dmat3x2: GL.GL_DOUBLE, glm.dmat3x3: GL.GL_DOUBLE, glm.dmat3x4: GL.GL_DOUBLE,
		glm.dmat4x2: GL.GL_DOUBLE, glm.dmat4x3: GL.GL_DOUBLE, glm.dmat4x4: GL.GL_DOUBLE,
	}

	np_dtype_map = \
	{
		int: np.int32, float: np.float32,
		glm.vec2: np.float32, glm.vec3: np.float32, glm.vec4: np.float32,
		glm.bvec2: np.int8, glm.bvec3: np.int8, glm.bvec4: np.int8,
		glm.ivec2: np.int32, glm.ivec3: np.int32, glm.ivec4: np.int32,
		glm.uvec2: np.uint32, glm.uvec3: np.uint32, glm.uvec4: np.uint32,
		glm.dvec2: np.float64, glm.dvec3: np.float64, glm.dvec4: np.float64,
		glm.mat2x2: np.float32, glm.mat2x3: np.float32, glm.mat2x4: np.float32,
		glm.mat3x2: np.float32, glm.mat3x3: np.float32, glm.mat3x4: np.float32,
		glm.mat4x2: np.float32, glm.mat4x3: np.float32, glm.mat4x4: np.float32,
		glm.dmat2x2: np.float64, glm.dmat2x3: np.float64, glm.dmat2x4: np.float64,
		glm.dmat3x2: np.float64, glm.dmat3x3: np.float64, glm.dmat3x4: np.float64,
		glm.dmat4x2: np.float64, glm.dmat4x3: np.float64, glm.dmat4x4: np.float64,
	}

	blend_equation_map = \
	{
		GL.GL_FUNC_ADD: "+",
		GL.GL_FUNC_SUBTRACT: "-",
		GL.GL_FUNC_REVERSE_SUBTRACT: GL.GL_FUNC_REVERSE_SUBTRACT,

		"+": GL.GL_FUNC_ADD,
		"-": GL.GL_FUNC_SUBTRACT
	}

	depth_func_map = \
	{
		GL.GL_ALWAYS: "always",
		GL.GL_NEVER: "never",
		GL.GL_LESS: "<",
		GL.GL_EQUAL: "==",
		GL.GL_LEQUAL: "<=",
		GL.GL_GREATER: ">",
		GL.GL_NOTEQUAL: "!=",
		GL.GL_GEQUAL: ">=",

		"always": GL.GL_ALWAYS,
		"never": GL.GL_NEVER,
		"<": GL.GL_LESS,
		"==": GL.GL_EQUAL,
		"<=": GL.GL_LEQUAL,
		">": GL.GL_GREATER,
		"!=": GL.GL_NOTEQUAL,
		">=": GL.GL_GEQUAL
	}

	stencil_func_map = \
	{
		GL.GL_ALWAYS: "always",
		GL.GL_NEVER: "never",
		GL.GL_LESS: "<",
		GL.GL_EQUAL: "==",
		GL.GL_LEQUAL: "<=",
		GL.GL_GREATER: ">",
		GL.GL_NOTEQUAL: "!=",
		GL.GL_GEQUAL: ">=",

		"always": GL.GL_ALWAYS,
		"never": GL.GL_NEVER,
		"<": GL.GL_LESS,
		"==": GL.GL_EQUAL,
		"<=": GL.GL_LEQUAL,
		">": GL.GL_GREATER,
		"!=": GL.GL_NOTEQUAL,
		">=": GL.GL_GEQUAL
	}

	internal_formats_map = \
	{
		np.uint8: {1: GL.GL_R8, 2: GL.GL_RG8, 3: GL.GL_RGB8, 4: GL.GL_RGBA8},
		np.int8: {1: GL.GL_R8I, 2: GL.GL_RG8I, 3: GL.GL_RGB8I, 4: GL.GL_RGBA8I},
		np.uint16: {1: GL.GL_R16, 2: GL.GL_RG16, 3: GL.GL_RGB16, 4: GL.GL_RGBA16},
		np.int16: {1: GL.GL_R16I, 2: GL.GL_RG16I, 3: GL.GL_RGB16I, 4: GL.GL_RGBA16I},
		np.uint32: {1: GL.GL_R32UI, 2: GL.GL_RG32UI, 3: GL.GL_RGB32UI, 4: GL.GL_RGBA32UI},
		np.int32: {1: GL.GL_R32I, 2: GL.GL_RG32I, 3: GL.GL_RGB32I, 4: GL.GL_RGBA32I},
		np.float16: {1: GL.GL_R16F, 2: GL.GL_RG16F, 3: GL.GL_RGB16F, 4: GL.GL_RGBA16F},
		np.float32: {1: GL.GL_R32F, 2: GL.GL_RG32F, 3: GL.GL_RGB32F, 4: GL.GL_RGBA32F},
		# np.float64: {1: GL.GL_R64, 2: GL.GL_RG64F, 3: GL.GL_RGB64F, 4: GL.GL_RGBA64F}

		dtype_uint8: {1: GL.GL_R8, 2: GL.GL_RG8, 3: GL.GL_RGB8, 4: GL.GL_RGBA8},
		dtype_int8: {1: GL.GL_R8I, 2: GL.GL_RG8I, 3: GL.GL_RGB8I, 4: GL.GL_RGBA8I},
		dtype_uint16: {1: GL.GL_R16, 2: GL.GL_RG16, 3: GL.GL_RGB16, 4: GL.GL_RGBA16},
		dtype_int16: {1: GL.GL_R16I, 2: GL.GL_RG16I, 3: GL.GL_RGB16I, 4: GL.GL_RGBA16I},
		dtype_uint32: {1: GL.GL_R32UI, 2: GL.GL_RG32UI, 3: GL.GL_RGB32UI, 4: GL.GL_RGBA32UI},
		dtype_int32: {1: GL.GL_R32I, 2: GL.GL_RG32I, 3: GL.GL_RGB32I, 4: GL.GL_RGBA32I},
		dtype_float16: {1: GL.GL_R16F, 2: GL.GL_RG16F, 3: GL.GL_RGB16F, 4: GL.GL_RGBA16F},
		dtype_float32: {1: GL.GL_R32F, 2: GL.GL_RG32F, 3: GL.GL_RGB32F, 4: GL.GL_RGBA32F},
		# dtype_float64: {1: GL.GL_R64, 2: GL.GL_RG64F, 3: GL.GL_RGB64F, 4: GL.GL_RGBA64F}
	}

	format_info_map = \
	{
		GL.GL_RED: [GL.GL_RED, GL.GL_UNSIGNED_BYTE],
		GL.GL_R8: [GL.GL_RED, GL.GL_UNSIGNED_BYTE],
		GL.GL_R8_SNORM: [GL.GL_RED, GL.GL_BYTE],
		GL.GL_R16: [GL.GL_RED, GL.GL_UNSIGNED_SHORT],
		GL.GL_R16_SNORM: [GL.GL_RED, GL.GL_SHORT],
		GL.GL_R16F: [GL.GL_RED, GL.GL_HALF_FLOAT],
		GL.GL_R32F: [GL.GL_RED, GL.GL_FLOAT],
		GL.GL_R8I: [GL.GL_RED_INTEGER, GL.GL_BYTE],
		GL.GL_R8UI: [GL.GL_RED_INTEGER, GL.GL_UNSIGNED_BYTE],
		GL.GL_R16I: [GL.GL_RED_INTEGER, GL.GL_SHORT],
		GL.GL_R16UI: [GL.GL_RED_INTEGER, GL.GL_UNSIGNED_SHORT],
		GL.GL_R32I: [GL.GL_RED_INTEGER, GL.GL_INT],
		GL.GL_R32UI: [GL.GL_RED_INTEGER, GL.GL_UNSIGNED_INT],
		GL.GL_COMPRESSED_RED: [GL.GL_RED, GL.GL_UNSIGNED_BYTE],
		GL.GL_COMPRESSED_RED_RGTC1: [GL.GL_RED, GL.GL_UNSIGNED_BYTE],
		GL.GL_COMPRESSED_SIGNED_RED_RGTC1: [GL.GL_RED, GL.GL_BYTE],
		GL.GL_RG: [GL.GL_RG, GL.GL_UNSIGNED_BYTE],
		GL.GL_RG8: [GL.GL_RG, GL.GL_UNSIGNED_BYTE],
		GL.GL_RG8_SNORM: [GL.GL_RG, GL.GL_BYTE],
		GL.GL_RG16: [GL.GL_RG, GL.GL_UNSIGNED_SHORT],
		GL.GL_RG16_SNORM: [GL.GL_RG, GL.GL_SHORT],
		GL.GL_RG16F: [GL.GL_RG, GL.GL_HALF_FLOAT],
		GL.GL_RG32F: [GL.GL_RG, GL.GL_FLOAT],
		GL.GL_RG8I: [GL.GL_RG_INTEGER, GL.GL_BYTE],
		GL.GL_RG8UI: [GL.GL_RG_INTEGER, GL.GL_UNSIGNED_BYTE],
		GL.GL_RG16I: [GL.GL_RG_INTEGER, GL.GL_SHORT],
		GL.GL_RG16UI: [GL.GL_RG_INTEGER, GL.GL_UNSIGNED_SHORT],
		GL.GL_RG32I: [GL.GL_RG_INTEGER, GL.GL_INT],
		GL.GL_RG32UI: [GL.GL_RG_INTEGER, GL.GL_UNSIGNED_INT],
		GL.GL_COMPRESSED_RG: [GL.GL_RG, GL.GL_UNSIGNED_BYTE],
		GL.GL_COMPRESSED_RG_RGTC2: [GL.GL_RG, GL.GL_UNSIGNED_BYTE],
		GL.GL_COMPRESSED_SIGNED_RG_RGTC2: [GL.GL_RG, GL.GL_BYTE],
		GL.GL_RGB: [GL.GL_RGB, GL.GL_UNSIGNED_BYTE],
		GL.GL_R3_G3_B2: [GL.GL_RGB, GL.GL_UNSIGNED_BYTE],
		GL.GL_RGB4: [GL.GL_RGB, GL.GL_UNSIGNED_BYTE],
		GL.GL_RGB5: [GL.GL_RGB, GL.GL_UNSIGNED_BYTE],
		GL.GL_RGB8: [GL.GL_RGB, GL.GL_UNSIGNED_BYTE],
		GL.GL_RGB8_SNORM: [GL.GL_RGB, GL.GL_BYTE],
		GL.GL_RGB10: [GL.GL_RGB, GL.GL_UNSIGNED_SHORT],
		GL.GL_RGB12: [GL.GL_RGB, GL.GL_UNSIGNED_SHORT],
		GL.GL_RGB16: [GL.GL_RGB, GL.GL_UNSIGNED_SHORT],
		GL.GL_RGB16_SNORM: [GL.GL_RGB, GL.GL_SHORT],
		GL.GL_SRGB8: [GL.GL_RGB, GL.GL_BYTE],
		GL.GL_RGB16F: [GL.GL_RGB, GL.GL_HALF_FLOAT],
		GL.GL_RGB32F: [GL.GL_RGB, GL.GL_FLOAT],
		GL.GL_R11F_G11F_B10F: [GL.GL_RGB, GL.GL_HALF_FLOAT],
		GL.GL_RGB9_E5: [GL.GL_RGB, GL.GL_SHORT],
		GL.GL_RGB8I: [GL.GL_RGB_INTEGER, GL.GL_BYTE],
		GL.GL_RGB8UI: [GL.GL_RGB_INTEGER, GL.GL_UNSIGNED_BYTE],
		GL.GL_RGB16I: [GL.GL_RGB_INTEGER, GL.GL_SHORT],
		GL.GL_RGB16UI: [GL.GL_RGB_INTEGER, GL.GL_UNSIGNED_SHORT],
		GL.GL_RGB32I: [GL.GL_RGB_INTEGER, GL.GL_INT],
		GL.GL_RGB32UI: [GL.GL_RGB_INTEGER, GL.GL_UNSIGNED_INT],
		GL.GL_COMPRESSED_RGB: [GL.GL_RGB, GL.GL_UNSIGNED_BYTE],
		GL.GL_COMPRESSED_SRGB: [GL.GL_RGB, GL.GL_UNSIGNED_BYTE],
		GL.GL_COMPRESSED_RGB_BPTC_SIGNED_FLOAT: [GL.GL_RGB, GL.GL_FLOAT],
		GL.GL_COMPRESSED_RGB_BPTC_UNSIGNED_FLOAT: [GL.GL_RGB, GL.GL_FLOAT],
		GL.GL_RGBA: [GL.GL_RGBA, GL.GL_UNSIGNED_BYTE],
		GL.GL_RGBA2: [GL.GL_RGBA, GL.GL_UNSIGNED_BYTE],
		GL.GL_RGBA4: [GL.GL_RGBA, GL.GL_UNSIGNED_BYTE],
		GL.GL_RGB5_A1: [GL.GL_RGBA, GL.GL_UNSIGNED_BYTE],
		GL.GL_RGBA8: [GL.GL_RGBA, GL.GL_UNSIGNED_BYTE],
		GL.GL_RGBA8_SNORM: [GL.GL_RGBA, GL.GL_BYTE],
		GL.GL_RGB10_A2: [GL.GL_RGBA, GL.GL_UNSIGNED_SHORT],
		GL.GL_RGB10_A2UI: [GL.GL_RGBA, GL.GL_UNSIGNED_SHORT],
		GL.GL_RGBA12: [GL.GL_RGBA, GL.GL_UNSIGNED_SHORT],
		GL.GL_RGBA16: [GL.GL_RGBA, GL.GL_UNSIGNED_SHORT],
		GL.GL_RGBA16_SNORM: [GL.GL_RGBA, GL.GL_UNSIGNED_SHORT],
		GL.GL_SRGB8_ALPHA8: [GL.GL_RGBA, GL.GL_UNSIGNED_BYTE],
		GL.GL_RGBA16F: [GL.GL_RGBA, GL.GL_HALF_FLOAT],
		GL.GL_RGBA32F: [GL.GL_RGBA, GL.GL_FLOAT],
		GL.GL_RGBA8I: [GL.GL_RGBA_INTEGER, GL.GL_BYTE],
		GL.GL_RGBA8UI: [GL.GL_RGBA_INTEGER, GL.GL_UNSIGNED_BYTE],
		GL.GL_RGBA16I: [GL.GL_RGBA_INTEGER, GL.GL_SHORT],
		GL.GL_RGBA16UI: [GL.GL_RGBA_INTEGER, GL.GL_UNSIGNED_SHORT],
		GL.GL_RGBA32I: [GL.GL_RGBA_INTEGER, GL.GL_INT],
		GL.GL_RGBA32UI: [GL.GL_RGBA_INTEGER, GL.GL_UNSIGNED_INT],
		GL.GL_COMPRESSED_RGBA: [GL.GL_RGBA, GL.GL_UNSIGNED_BYTE],
		GL.GL_COMPRESSED_SRGB_ALPHA: [GL.GL_RGBA, GL.GL_UNSIGNED_BYTE],
		GL.GL_COMPRESSED_RGBA_BPTC_UNORM: [GL.GL_RGBA, GL.GL_UNSIGNED_BYTE],
		GL.GL_COMPRESSED_SRGB_ALPHA_BPTC_UNORM: [GL.GL_RGBA, GL.GL_UNSIGNED_BYTE],

		GL.GL_DEPTH_COMPONENT: [GL.GL_DEPTH_COMPONENT, GL.GL_FLOAT],
		GL.GL_DEPTH_COMPONENT16: [GL.GL_DEPTH_COMPONENT, GL.GL_SHORT],
		GL.GL_DEPTH_COMPONENT24: [GL.GL_DEPTH_COMPONENT, GL.GL_INT],
		GL.GL_DEPTH_COMPONENT32: [GL.GL_DEPTH_COMPONENT, GL.GL_INT],
		GL.GL_DEPTH_COMPONENT32F: [GL.GL_DEPTH_COMPONENT, GL.GL_FLOAT],

		GL.GL_STENCIL_INDEX: [GL.GL_STENCIL_INDEX, GL.GL_UNSIGNED_BYTE],
		GL.GL_STENCIL_INDEX1: [GL.GL_STENCIL_INDEX, GL.GL_UNSIGNED_BYTE],
		GL.GL_STENCIL_INDEX4: [GL.GL_STENCIL_INDEX, GL.GL_UNSIGNED_BYTE],
		GL.GL_STENCIL_INDEX8: [GL.GL_STENCIL_INDEX, GL.GL_UNSIGNED_BYTE],
		GL.GL_STENCIL_INDEX16: [GL.GL_STENCIL_INDEX, GL.GL_UNSIGNED_SHORT],

		GL.GL_DEPTH_STENCIL: [GL.GL_DEPTH_STENCIL, GL.GL_FLOAT],
		GL.GL_DEPTH24_STENCIL8: [GL.GL_DEPTH_STENCIL, GL.GL_FLOAT],
		GL.GL_DEPTH32F_STENCIL8: [GL.GL_DEPTH_STENCIL, GL.GL_FLOAT]
	}

	enum_map = {value:value for name, value in vars(GL).items() if name.startswith('GL_') and isinstance(value, constant.IntConstant)}

	internal_formats = \
	[
		GL.GL_RED, GL.GL_R8, GL.GL_R8_SNORM, GL.GL_R16, GL.GL_R16_SNORM,
		GL.GL_R16F, GL.GL_R32F, GL.GL_R8I, GL.GL_R8UI, GL.GL_R16I, GL.GL_R16UI,
		GL.GL_R32I, GL.GL_R32UI, GL.GL_COMPRESSED_RED, GL.GL_COMPRESSED_RED_RGTC1,
		GL.GL_COMPRESSED_SIGNED_RED_RGTC1, GL.GL_RG, GL.GL_RG8, GL.GL_RG8_SNORM, GL.GL_RG16,
		GL.GL_RG16_SNORM, GL.GL_RG16F, GL.GL_RG32F, GL.GL_RG8I, GL.GL_RG8UI, GL.GL_RG16I,
		GL.GL_RG16UI, GL.GL_RG32I, GL.GL_RG32UI, GL.GL_COMPRESSED_RG, GL.GL_COMPRESSED_RG_RGTC2,
		GL.GL_COMPRESSED_SIGNED_RG_RGTC2, GL.GL_RGB, GL.GL_R3_G3_B2, GL.GL_RGB4, GL.GL_RGB5,
		GL.GL_RGB8, GL.GL_RGB8_SNORM, GL.GL_RGB10, GL.GL_RGB12, GL.GL_RGB16, GL.GL_RGB16_SNORM, GL.GL_SRGB8,
		GL.GL_RGB16F, GL.GL_RGB32F, GL.GL_R11F_G11F_B10F, GL.GL_RGB9_E5, GL.GL_RGB8I,
		GL.GL_RGB8UI, GL.GL_RGB16I, GL.GL_RGB16UI, GL.GL_RGB32I, GL.GL_RGB32UI,
		GL.GL_COMPRESSED_RGB, GL.GL_COMPRESSED_SRGB, GL.GL_COMPRESSED_RGB_BPTC_SIGNED_FLOAT,
		GL.GL_COMPRESSED_RGB_BPTC_UNSIGNED_FLOAT, GL.GL_RGBA, GL.GL_RGBA2, GL.GL_RGBA4,
		GL.GL_RGB5_A1, GL.GL_RGBA8, GL.GL_RGBA8_SNORM, GL.GL_RGB10_A2, GL.GL_RGB10_A2UI,
		GL.GL_RGBA12, GL.GL_RGBA16, GL.GL_RGBA16_SNORM, GL.GL_SRGB8_ALPHA8, GL.GL_RGBA16F,
		GL.GL_RGBA32F, GL.GL_RGBA8I, GL.GL_RGBA8UI, GL.GL_RGBA16I, GL.GL_RGBA16UI,
		GL.GL_RGBA32I, GL.GL_RGBA32UI, GL.GL_COMPRESSED_RGBA, GL.GL_COMPRESSED_SRGB_ALPHA,
		GL.GL_COMPRESSED_RGBA_BPTC_UNORM, GL.GL_COMPRESSED_SRGB_ALPHA_BPTC_UNORM,

		GL.GL_DEPTH_COMPONENT, GL.GL_DEPTH_COMPONENT16, GL.GL_DEPTH_COMPONENT24, GL.GL_DEPTH_COMPONENT32, GL.GL_DEPTH_COMPONENT32F,
		GL.GL_STENCIL_INDEX, GL.GL_STENCIL_INDEX1, GL.GL_STENCIL_INDEX4, GL.GL_STENCIL_INDEX8, GL.GL_STENCIL_INDEX16,
		GL.GL_DEPTH_STENCIL, GL.GL_DEPTH24_STENCIL8, GL.GL_DEPTH32F_STENCIL8, None
	]

	color_internal_formats = \
	[
		GL.GL_RED, GL.GL_R8, GL.GL_R8_SNORM, GL.GL_R16, GL.GL_R16_SNORM,
		GL.GL_R16F, GL.GL_R32F, GL.GL_R8I, GL.GL_R8UI, GL.GL_R16I, GL.GL_R16UI,
		GL.GL_R32I, GL.GL_R32UI, GL.GL_COMPRESSED_RED, GL.GL_COMPRESSED_RED_RGTC1,
		GL.GL_COMPRESSED_SIGNED_RED_RGTC1, GL.GL_RG, GL.GL_RG8, GL.GL_RG8_SNORM, GL.GL_RG16,
		GL.GL_RG16_SNORM, GL.GL_RG16F, GL.GL_RG32F, GL.GL_RG8I, GL.GL_RG8UI, GL.GL_RG16I,
		GL.GL_RG16UI, GL.GL_RG32I, GL.GL_RG32UI, GL.GL_COMPRESSED_RG, GL.GL_COMPRESSED_RG_RGTC2,
		GL.GL_COMPRESSED_SIGNED_RG_RGTC2, GL.GL_RGB, GL.GL_R3_G3_B2, GL.GL_RGB4, GL.GL_RGB5,
		GL.GL_RGB8, GL.GL_RGB8_SNORM, GL.GL_RGB10, GL.GL_RGB12, GL.GL_RGB16, GL.GL_RGB16_SNORM, GL.GL_SRGB8,
		GL.GL_RGB16F, GL.GL_RGB32F, GL.GL_R11F_G11F_B10F, GL.GL_RGB9_E5, GL.GL_RGB8I,
		GL.GL_RGB8UI, GL.GL_RGB16I, GL.GL_RGB16UI, GL.GL_RGB32I, GL.GL_RGB32UI,
		GL.GL_COMPRESSED_RGB, GL.GL_COMPRESSED_SRGB, GL.GL_COMPRESSED_RGB_BPTC_SIGNED_FLOAT,
		GL.GL_COMPRESSED_RGB_BPTC_UNSIGNED_FLOAT, GL.GL_RGBA, GL.GL_RGBA2, GL.GL_RGBA4,
		GL.GL_RGB5_A1, GL.GL_RGBA8, GL.GL_RGBA8_SNORM, GL.GL_RGB10_A2, GL.GL_RGB10_A2UI,
		GL.GL_RGBA12, GL.GL_RGBA16, GL.GL_RGBA16_SNORM, GL.GL_SRGB8_ALPHA8, GL.GL_RGBA16F,
		GL.GL_RGBA32F, GL.GL_RGBA8I, GL.GL_RGBA8UI, GL.GL_RGBA16I, GL.GL_RGBA16UI,
		GL.GL_RGBA32I, GL.GL_RGBA32UI, GL.GL_COMPRESSED_RGBA, GL.GL_COMPRESSED_SRGB_ALPHA,
		GL.GL_COMPRESSED_RGBA_BPTC_UNORM, GL.GL_COMPRESSED_SRGB_ALPHA_BPTC_UNORM
	]

	image_internal_formats = \
	[
		GL.GL_RGBA32F, GL.GL_RGBA16F, GL.GL_RG32F, GL.GL_RG16F, GL.GL_R11F_G11F_B10F,
		GL.GL_R32F, GL.GL_R16F, GL.GL_RGBA16, GL.GL_RGB10_A2, GL.GL_RGBA8, GL.GL_RG16,
		GL.GL_RG8, GL.GL_R16, GL.GL_R8, GL.GL_RGBA16_SNORM, GL.GL_RGBA8_SNORM,
		GL.GL_RG16_SNORM, GL.GL_RG8_SNORM, GL.GL_R16_SNORM, None
	]

	isampler_internal_formats = \
	[
		GL.GL_RGBA32I, GL.GL_RGBA16I, GL.GL_RGBA8I,
		GL.GL_RGB32I, GL.GL_RGB16I, GL.GL_RGB8I,
		GL.GL_RG32I, GL.GL_RG16I, GL.GL_RG8I,
		GL.GL_R32I, GL.GL_R16I, GL.GL_R8I, None
	]

	iimage_internal_formats = \
	[
		GL.GL_RGBA32I, GL.GL_RGBA16I, GL.GL_RGBA8I, GL.GL_RG32I, GL.GL_RG16I,
		GL.GL_RG8I, GL.GL_R32I, GL.GL_R16I, GL.GL_R8I, None
	]

	usampler_internal_formats = \
	[
		GL.GL_RGBA32UI, GL.GL_RGBA16UI, GL.GL_RGB10_A2UI, GL.GL_RGBA8UI,
		GL.GL_RGB32UI, GL.GL_RGB16UI, GL.GL_RGB8UI,
		GL.GL_RG32UI, GL.GL_RG16UI, GL.GL_RG8UI,
		GL.GL_R32UI, GL.GL_R16UI, GL.GL_R8UI, None
	]

	uimage_internal_formats = \
	[
		GL.GL_RGBA32UI, GL.GL_RGBA16UI, GL.GL_RGB10_A2UI, GL.GL_RGBA8UI,
		GL.GL_RG32UI, GL.GL_RG16UI, GL.GL_RG8UI, GL.GL_R32UI, GL.GL_R16UI,
		GL.GL_R8UI, None
	]

	depth_internal_formats = \
	[
		GL.GL_DEPTH_COMPONENT, GL.GL_DEPTH_COMPONENT16, GL.GL_DEPTH_COMPONENT24,
		GL.GL_DEPTH_COMPONENT32, GL.GL_DEPTH_COMPONENT32F,
	]

	stencil_internal_formats = \
	[
		GL.GL_STENCIL_INDEX, GL.GL_STENCIL_INDEX1, GL.GL_STENCIL_INDEX4,
		GL.GL_STENCIL_INDEX8, GL.GL_STENCIL_INDEX16,
	]

	depth_stencil_internal_formats = \
	[
		GL.GL_DEPTH_STENCIL, GL.GL_DEPTH24_STENCIL8, GL.GL_DEPTH32F_STENCIL8
	]

	fog_modes = [GL.GL_LINEAR, GL.GL_EXP, GL.GL_EXP2]

	dtypes = \
	[
		GL.GL_BYTE, GL.GL_UNSIGNED_BYTE, GL.GL_SHORT, GL.GL_UNSIGNED_SHORT, GL.GL_INT,
		GL.GL_UNSIGNED_INT, GL.GL_HALF_FLOAT, GL.GL_FLOAT, GL.GL_DOUBLE, None
	]

	operations = \
	[
		GL.GL_KEEP, # 保持模板缓冲区的当前值
		GL.GL_ZERO, # 将模板缓冲区的值设置为 0
		GL.GL_REPLACE, # 将模板缓冲区的值替换为参考值（通过 glStencilFunc() 设置）
		GL.GL_INCR, # 将模板缓冲区的值递增 1（如果结果超出了模板缓冲区的范围，将会回环到 0）
		GL.GL_DECR, # 将模板缓冲区的值递减 1（如果结果小于 0，将会回环到最大值）
		GL.GL_INVERT # 按位取反模板缓冲区的值
	]

	shader_ext_map = \
	{
		".vert": GL.GL_VERTEX_SHADER,
		".tesc": GL.GL_TESS_CONTROL_SHADER,
		".tese": GL.GL_TESS_EVALUATION_SHADER,
		".geom": GL.GL_GEOMETRY_SHADER,
		".frag": GL.GL_FRAGMENT_SHADER,
		".comp": GL.GL_COMPUTE_SHADER,
		
		".vs": GL.GL_VERTEX_SHADER,
		".tcs": GL.GL_TESS_CONTROL_SHADER,
		".tes": GL.GL_TESS_EVALUATION_SHADER,
		".gs": GL.GL_GEOMETRY_SHADER,
		".fs": GL.GL_FRAGMENT_SHADER,
		".cs": GL.GL_COMPUTE_SHADER,

		GL.GL_VERTEX_SHADER: ".vert",
		GL.GL_TESS_CONTROL_SHADER: ".tesc",
		GL.GL_TESS_EVALUATION_SHADER: ".tese",
		GL.GL_GEOMETRY_SHADER: ".geom",
		GL.GL_FRAGMENT_SHADER: ".frag",
		GL.GL_COMPUTE_SHADER: ".comp"
	}

	primitive_type_map = \
	{
		"points": [GL.GL_POINTS],
		"lines": [GL.GL_LINES, GL.GL_LINE_STRIP, GL.GL_LINE_LOOP],
		"lines_adjacency": [GL.GL_LINES_ADJACENCY, GL.GL_LINE_STRIP_ADJACENCY],
		"triangles": [GL.GL_TRIANGLES, GL.GL_TRIANGLE_STRIP, GL.GL_TRIANGLE_FAN],
		"triangles_adjacency": [GL.GL_TRIANGLES_ADJACENCY, GL.GL_TRIANGLE_STRIP_ADJACENCY]
	}

	polygon_modes = [GL.GL_FILL, GL.GL_LINE, GL.GL_POINT]

	depth_funcs = [GL.GL_ALWAYS, GL.GL_NEVER, GL.GL_LESS, GL.GL_EQUAL, GL.GL_LEQUAL, GL.GL_GREATER, GL.GL_NOTEQUAL, GL.GL_GEQUAL]

	depth_func_strs = ["always", "never", "<", "==", "<=", ">", "!=", ">="]

	stencil_funcs = [GL.GL_ALWAYS, GL.GL_NEVER, GL.GL_LESS, GL.GL_EQUAL, GL.GL_LEQUAL, GL.GL_GREATER, GL.GL_NOTEQUAL, GL.GL_GEQUAL]

	stencil_func_strs = ["always", "never", "<", "==", "<=", ">", "!=", ">="]

	shader_types = [GL.GL_VERTEX_SHADER, GL.GL_FRAGMENT_SHADER, GL.GL_GEOMETRY_SHADER, GL.GL_TESS_CONTROL_SHADER, GL.GL_TESS_EVALUATION_SHADER, None]
	
	wrap_types = [GL.GL_REPEAT, GL.GL_MIRRORED_REPEAT, GL.GL_CLAMP_TO_EDGE, GL.GL_CLAMP_TO_BORDER, GL.GL_MIRROR_CLAMP_TO_EDGE]
	
	filter_types = [GL.GL_NEAREST, GL.GL_LINEAR, None]
	
	none_color_attachment_types = [GL.GL_DEPTH_ATTACHMENT, GL.GL_STENCIL_ATTACHMENT, GL.GL_DEPTH_STENCIL_ATTACHMENT]
	
	draw_types = [GL.GL_STATIC_DRAW, GL.GL_DYNAMIC_DRAW, GL.GL_STREAM_DRAW, GL.GL_DYNAMIC_COPY]
	
	triangle_types = [GL.GL_TRIANGLES, GL.GL_TRIANGLE_STRIP, GL.GL_TRIANGLE_FAN]
	
	line_types = [GL.GL_LINES, GL.GL_LINE_LOOP, GL.GL_LINE_STRIP]
	
	primitive_types = [*triangle_types, *line_types, GL.GL_POINTS, GL.GL_PATCHES]
	
	cull_face_types = [GL.GL_BACK, GL.GL_FRONT, GL.GL_FRONT_AND_BACK, None]
	
	blend_equations = [GL.GL_FUNC_ADD, GL.GL_FUNC_SUBTRACT, GL.GL_FUNC_REVERSE_SUBTRACT]
	
	blend_funcs = \
	[
		GL.GL_ZERO, GL.GL_ONE,
		GL.GL_SRC_COLOR, GL.GL_ONE_MINUS_SRC_COLOR,
		GL.GL_DST_COLOR, GL.GL_ONE_MINUS_DST_COLOR,
		GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA,
		GL.GL_DST_ALPHA, GL.GL_ONE_MINUS_DST_ALPHA,
		GL.GL_CONSTANT_COLOR, GL.GL_ONE_MINUS_CONSTANT_COLOR,
		GL.GL_CONSTANT_ALPHA, GL.GL_ONE_MINUS_CONSTANT_ALPHA
	]

	attr_types = \
	(
		int, float,
		glm.vec2, glm.ivec2, glm.uvec2, glm.dvec2,
		glm.vec3, glm.ivec3, glm.uvec3, glm.dvec3,
		glm.vec4, glm.ivec4, glm.uvec4, glm.dvec4,

		glm.mat2x2, glm.mat2x3, glm.mat2x4,
		glm.mat3x2, glm.mat3x3, glm.mat3x4,
		glm.mat4x2, glm.mat4x3, glm.mat4x4,

		glm.dmat2x2, glm.dmat2x3, glm.dmat2x4,
		glm.dmat3x2, glm.dmat3x3, glm.dmat3x4,
		glm.dmat4x2, glm.dmat4x3, glm.dmat4x4,
	)

	int_types = [GL.GL_BYTE, GL.GL_UNSIGNED_BYTE, GL.GL_SHORT, GL.GL_UNSIGNED_SHORT, GL.GL_INT, GL.GL_UNSIGNED_INT]

	atom_type_names = \
	[
		"bool", "int", "uint", "uint64_t", "float", "double", "atomic_uint",

		"bvec2", "bvec3", "bvec4",
		"ivec2", "ivec3", "ivec4",
		"uvec2", "uvec3", "uvec4",
		"vec2", "vec3", "vec4",
		"dvec2", "dvec3", "dvec4",

		"mat2x2", "mat3x2", "mat4x2",
		"mat2x3", "mat3x3", "mat4x3",
		"mat2x4", "mat3x4", "mat4x4",
		"mat2", "mat3", "mat4",

		"dmat2x2", "dmat3x2", "dmat4x2",
		"dmat2x3", "dmat3x3", "dmat4x3",
		"dmat2x4", "dmat3x4", "dmat4x4",
		"dmat2", "dmat3", "dmat4",

		"sampler2D", "isampler2D", "usampler2D",
		"sampler2DMS", "isampler2DMS", "usampler2DMS",
		"sampler2DArray", "isampler2DArray", "usampler2DArray",
		"sampler2DMSArray", "isampler2DMSArray", "usampler2DMSArray",
		"samplerCube", "samplerCubeArray",
		"image2D", "iimage2D", "uimage2D"
	]

	atom_type_map = \
	{
		"bool":bool, "int":int, "uint":np.uint32, "uint64_t":np.uint64,
		"float":np.float32, "double":np.float64, "atomic_uint":np.uint32,

		"bvec2":glm.bvec2, "bvec3":glm.bvec3, "bvec4":glm.bvec4,
		"ivec2":glm.ivec2, "ivec3":glm.ivec3, "ivec4":glm.ivec4,
		"uvec2":glm.uvec2, "uvec3":glm.uvec3, "uvec4":glm.uvec4,
		"vec2":glm.vec2, "vec3":glm.vec3, "vec4":glm.vec4,
		"dvec2":glm.dvec2, "dvec3":glm.dvec3, "dvec4":glm.dvec4,

		"mat2x2":glm.mat2x2, "mat3x2":glm.mat3x2, "mat4x2":glm.mat4x2,
		"mat2x3":glm.mat2x3, "mat3x3":glm.mat3x3, "mat4x3":glm.mat4x3,
		"mat2x4":glm.mat2x4, "mat3x4":glm.mat3x4, "mat4x4":glm.mat4x4,
		"mat2":glm.mat2, "mat3":glm.mat3, "mat4":glm.mat4,

		"dmat2x2":glm.dmat2x2, "dmat3x2":glm.dmat3x2, "dmat4x2":glm.dmat4x2,
		"dmat2x3":glm.dmat2x3, "dmat3x3":glm.dmat3x3, "dmat4x3":glm.dmat4x3,
		"dmat2x4":glm.dmat2x4, "dmat3x4":glm.dmat3x4, "dmat4x4":glm.dmat4x4,
		"dmat2":glm.dmat2, "dmat3":glm.dmat3, "dmat4":glm.dmat4
	}

	memory_modifiers_to_internal_types_map = \
	{
		"r11f_g11f_b10f": GL.GL_R11F_G11F_B10F,
		"rgba16_snorm": GL.GL_RGBA16_SNORM,
		"rgba8_snorm": GL.GL_RGBA8_SNORM,
		"rg16_snorm": GL.GL_RG16_SNORM,
		"rgb10_a2ui": GL.GL_RGB10_A2UI,
		"rg8_snorm": GL.GL_RG8_SNORM,
		"r16_snorm": GL.GL_R16_SNORM,
		"r8_snorm": GL.GL_R8_SNORM,
		"rgb10_a2": GL.GL_RGB10_A2,
		"rgba32ui": GL.GL_RGBA32UI,
		"rgba16ui": GL.GL_RGBA16UI,
		"rgba8ui": GL.GL_RGBA8UI,
		"rgba32f": GL.GL_RGBA32F,
		"rgba16f": GL.GL_RGBA16F,
		"rgba32i": GL.GL_RGBA32UI,
		"rgba16i": GL.GL_RGBA16I,
		"rgba16": GL.GL_RGBA16,
		"rgba8i": GL.GL_RGBA8I,
		"rg32ui": GL.GL_RG32UI,
		"rg16ui": GL.GL_RG16UI,
		"rg8ui": GL.GL_RG8UI,
		"r32ui": GL.GL_R32UI,
		"r16ui": GL.GL_R16UI,
		"rg32i": GL.GL_RG32I,
		"rg16i": GL.GL_RG16I,
		"rg32f": GL.GL_RG32F,
		"rg16f": GL.GL_RG16F,
		"rgba8": GL.GL_RGBA8,
		"rg8i": GL.GL_RG8I,
		"r32i": GL.GL_R32I,
		"r16i": GL.GL_R16I,
		"r32f": GL.GL_R32F,
		"r16f": GL.GL_R16F,
		"rg16": GL.GL_RG16,
		"r8ui": GL.GL_R8UI,
		"r8i": GL.GL_R8I,
		"rg8": GL.GL_RG8,
		"r16": GL.GL_R16,
		"r8": GL.GL_R8
	}
	mipmap_filter_map = \
	{
		(GL.GL_NEAREST, GL.GL_NEAREST): GL.GL_NEAREST_MIPMAP_NEAREST,
		(GL.GL_NEAREST, GL.GL_LINEAR): GL.GL_NEAREST_MIPMAP_LINEAR,
		(GL.GL_LINEAR, GL.GL_NEAREST): GL.GL_LINEAR_MIPMAP_NEAREST,
		(GL.GL_LINEAR, GL.GL_LINEAR): GL.GL_LINEAR_MIPMAP_LINEAR
	}
	fbo_status_errors = \
	{
		GL.GL_FRAMEBUFFER_UNDEFINED: "default FBO does not exist",
		GL.GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT: "some attachments are incomplete",
		GL.GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT: "no attachments attached",
		GL.GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER: "all color attachments' GL_FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE is GL_NONE",
		GL.GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER: "all read buffer color attachments' GL_FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE is GL_NONE",
		GL.GL_FRAMEBUFFER_UNSUPPORTED: "the combination of attachments' internal formats violates an implementation-dependent set of restrictions",
		GL.GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE: 
		[
			"GL_RENDERBUFFER_SAMPLES is not the same for all attached RBOs",
			"GL_TEXTURE_SAMPLES is the not same for all attached textures",
			"attachments are a mix of RBOs and textures, but GL_RENDERBUFFER_SAMPLES doesn't match GL_TEXTURE_SAMPLES",
			"GL_TEXTURE_FIXED_SAMPLE_LOCATIONS is not the same for all attached textures",
			"attachments are a mix of RBOs and textures, but GL_TEXTURE_FIXED_SAMPLE_LOCATIONS is not GL_TRUE for all attached textures",
		],
		GL.GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS:
		[
			"some attachments are layered, but some populated attachments are not layered",
			"all populated color attachments are not from textures of the same target"
		]
	}


from PyQt6.QtWidgets import QWidget, QHBoxLayout, QCheckBox
from PyQt6.QtCore import pyqtSignal, Qt
from .ImageChooser import ImageChooser

class MapInput(QWidget):

    checked_changed = pyqtSignal(bool)
    file_path_changed = pyqtSignal(str)

    def __init__(self, prompt:str="贴图", file_path:str="", width:int=50, height:int=50, parent:QWidget=None):
        QWidget.__init__(self, parent=parent)
        hlayout = QHBoxLayout()
        hlayout.setAlignment(Qt.AlignmentFlag.AlignLeft)

        self.__prompt_checkbox = QCheckBox(prompt, self)
        self.__image_chooser = ImageChooser(file_path, width, height, self)
        self.__image_chooser.setEnabled(False)
        hlayout.addWidget(self.__prompt_checkbox)
        hlayout.addWidget(self.__image_chooser)
        hlayout.setContentsMargins(0, 0, 0, 0)
        hlayout.setAlignment(Qt.AlignmentFlag.AlignLeft)
        self.setLayout(hlayout)

        self.__prompt_checkbox.stateChanged.connect(self.__checkbox_state_changed)
        self.__image_chooser.file_path_changed.connect(self.file_path_changed)

    def __checkbox_state_changed(self, state:Qt.CheckState):
        self.__image_chooser.setEnabled(state == Qt.CheckState.Checked.value)
        self.__image_chooser.update()
        self.checked_changed.emit(state == Qt.CheckState.Checked.value)

    @property
    def checked(self):
        return self.__prompt_checkbox.isChecked()
    
    @checked.setter
    def checked(self, checked:bool):
        self.__prompt_checkbox.setChecked(checked)

    @property
    def file_path(self):
        return self.__image_chooser.file_path
    
    @file_path.setter
    def file_path(self, file_path:str):
        self.__image_chooser.file_path = file_path


from OpenGL import GL, constant
import numpy as np
import glm
import struct

from .GLInfo import GLInfo

def glGetInt(gl_constant:constant.IntConstant)->int:
	value = GL.glGetIntegerv(gl_constant)
	try:
		value = value[0]
	except:
		pass

	value = int(value)
	return value

def glGetEnum(gl_constant:constant.IntConstant):
	value = glGetInt(gl_constant)
	if value in GLInfo.enum_map:
		return GLInfo.enum_map[value]
	else:
		return value
	
def glGetEnumi(gl_constant:constant.IntConstant, index:int):
	value = int(GL.glGetIntegeri_v(gl_constant, index))
	if value in GLInfo.enum_map:
		return GLInfo.enum_map[value]
	else:
		return value

def width_adapt(width):
	if width % 4 != 0:
		if width % 2 == 0:
			GL.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 2)
		else:
			GL.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1)

def sizeof(type_var):
	if not isinstance(type_var, type):
		type_var = type(type_var)

	if "sampler" in str(type_var):
		return 8

	size = 0

	try:
		size = glm.sizeof(type_var)
	except:
		pass

	if size != 0:
		return size

	if type_var == bool:
		return 1
	elif type_var in [int, float]:
		return 4

	return np.dtype(type_var).itemsize

def nitems(element_type):
	str_element_type = str(element_type)
	if 'vec2' in str_element_type:
		return 2
	elif 'vec3' in str_element_type:
		return 3
	elif 'vec4' in str_element_type or 'mat2x2' in str_element_type:
		return 4
	elif 'mat2x3' in str_element_type or 'mat3x2' in str_element_type:
		return 6
	elif 'mat3x3' in str_element_type:
		return 9
	elif 'mat2x4' in str_element_type or 'mat4x2' in str_element_type:
		return 8
	elif 'mat3x4' in str_element_type or 'mat4x3' in str_element_type:
		return 12
	elif 'mat4x4' in str_element_type:
		return 16
	else:
		return 1

def to_bytes(value):
	if isinstance(value, bool):
		return int(value).to_bytes(4, byteorder="little")
	elif isinstance(value, int):
		return value.to_bytes(4, byteorder="little")
	elif isinstance(value, float):
		return struct.pack('<f', value)
	else:
		return value.to_bytes()

def type_from_str(type_str):
	if type_str in GLInfo.atom_type_map:
		return GLInfo.atom_type_map[type_str]
	
	if type_str == "sampler2D":
		from .sampler2D import sampler2D
		return sampler2D
	elif type_str == "isampler2D":
		from .isampler2D import isampler2D
		return isampler2D
	elif type_str == "usampler2D":
		from .usampler2D import usampler2D
		return usampler2D
	elif type_str == "samplerCube":
		from .samplerCube import samplerCube
		return samplerCube
	elif type_str == "sampler2DMS":
		from .sampler2DMS import sampler2DMS
		return sampler2DMS
	elif type_str == "isampler2DMS":
		from .isampler2DMS import isampler2DMS
		return isampler2DMS
	elif type_str == "usampler2DMS":
		from .usampler2DMS import usampler2DMS
		return usampler2DMS
	else:
		raise TypeError(f"not support type {type_str}")

def get_external_format(internal_format):
	if internal_format not in GLInfo.format_info_map:
		return GL.GL_RGB
	else:
		return GLInfo.format_info_map[internal_format][0]

def get_channels(internal_format):
	external_format = get_external_format(internal_format)
	if external_format == GL.GL_RED: return 1
	elif external_format == GL.GL_RG: return 2
	elif external_format == GL.GL_RGB: return 3
	elif external_format == GL.GL_RGBA: return 4
	elif external_format == GL.GL_STENCIL_INDEX: return 1
	elif external_format == GL.GL_DEPTH_COMPONENT: return 1
	elif external_format == GL.GL_DEPTH_STENCIL: return 2
	else: return 3

def get_dtype(internal_format):
	if internal_format not in GLInfo.format_info_map:
		return GL.GL_UNSIGNED_BYTE
	else:
		return GLInfo.format_info_map[internal_format][1]



from .CommonRenderer import CommonRenderer
from glass import GLConfig, sampler2D

        
class ForwardRenderer(CommonRenderer):

    def __init__(self):
        CommonRenderer.__init__(self)

    def startup(self):
        CommonRenderer.startup(self)
        screen = self.camera.screen
        if not screen._samples_set_by_user and not screen._is_gl_init:
            screen._set_samples(4)

    def draw_opaque(self):
        if not self._opaque_meshes and \
           not self._opaque_lines and \
           not self._opaque_points:
            return

        with GLConfig.LocalConfig(depth_test=True, blend=False):
            with self.OIT_fbo:
                GLConfig.clear_buffers()

                if self._opaque_meshes:
                    self.prepare_forward_draw_mesh(True)
                    for mesh, instances in self._opaque_meshes:
                        self.forward_draw_mesh(mesh, instances)

                if self._opaque_lines:
                    self.prepare_forward_draw_lines(True)
                    for mesh, instances in self._opaque_lines:
                        self.forward_draw_lines(mesh, instances)

                if self._opaque_points:
                    self.prepare_forward_draw_points(True)
                    for mesh, instances in self._opaque_points:
                        self.forward_draw_points(mesh, instances)

                # 绘制天空盒
                if self.scene.skybox.is_completed:
                    self.scene.skybox.draw(self.camera)

                # 绘制天空穹顶
                elif self.scene.skydome.is_completed:
                    self.scene.skydome.draw(self.camera)

        self.OIT_fbo.draw_to_active(0)

        resolved = self.OIT_fbo.resolved
        self._depth_map = resolved.depth_attachment
        self._view_pos_map = resolved.color_attachment(3)
        self._view_normal_map = resolved.color_attachment(4)
    
    def render(self):
        # profiler.enable()
        self._should_update = False
        sampler2D._should_update = False
        self.classify_meshes()
        self.update_dir_lights_depth()
        self.update_point_lights_depth()
        self.update_spot_lights_depth()
        self.draw_opaque()
        self.draw_transparent()
        # profiler.disable()
        return (self._should_update or sampler2D._should_update)



from .Scene import Scene
from .SceneNode import SceneNode
from .BasicScene import SceneRoam, ModelView

from .Camera import Camera
from .Material import Material
from .Mesh import Mesh
from .Model import Model

from .ColorMap import ColorMap
from .Frame import Frame
from .Fog import Fog

import glm
from OpenGL import GL
from glass import GlassConfig



import setuptools
import os
import shutil

def find_files(module, directory):
    file_list = []
    for root, _, files in os.walk(directory):
        for file in files:
            file_path = os.path.join(root, file)[(len(module) + 1) :]
            if "__glcache__" not in file_path:
                file_list.append(file_path.replace("\\", "/"))
    return file_list

if os.path.isdir("build"):
    shutil.rmtree("build")

if os.path.isdir("glass_engine.egg-info"):
    shutil.rmtree("glass_engine.egg-info")

with open("glass_engine/README_PYPI.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

glass_engine_extra_files = \
[
    "images/glass_engine_logo64.png",
    "images/glass_engine_logo256.png",
    "images/start.png",
    "README_PYPI.md"
]
glass_engine_extra_files.extend(find_files("glass_engine", "glass_engine/glsl"))

glass_extra_files = \
[
    "tree-sitter-glsl/glsl.dll"
]
glass_extra_files.extend(find_files("glass", "glass/glsl"))

setuptools.setup(
    name="glass_engine",
    version="0.1.29",
    author="王炳辉 (BingHui-WANG)",
    author_email="binghui.wang@foxmail.com",
    description="An easy-to-use 3D rendering engine for Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Time-Coder/Glass-Engine",
    packages=setuptools.find_packages(),
    package_data={
        'glass_engine': glass_engine_extra_files,
        'glass': glass_extra_files
    },
    platforms=["win_amd64", "win32"],
    python_requires=">=3.6",
    install_requires=[
        "PyOpenGL",
        "PyOpenGL_accelerate",
        "moderngl",
        "PyGLM",
        "MarkupSafe==2.0.1",
        "qt-material",
        "numpy",
        "opencv-python",
        "Pillow",
        "maxminddb-geolite2",
        "wget",
        "requests",
        "chardet",
        "tree-sitter"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows",
    ],
)



from ..Mesh import Mesh

from glass.utils import checktype
from glass import Vertex

import glm

class RectFace(Mesh):

    @checktype
    def __init__(self, width:float=2, height:float=1,
                 color:(glm.vec3,glm.vec4)=glm.vec4(0.396, 0.74151, 0.69102, 1), back_color:(glm.vec3,glm.vec4)=None,
                 vertical:bool=False, normalize_tex_coord:bool=False, name:str=""):
        Mesh.__init__(self, color=color, back_color=back_color, name=name, block=True)
        self.__width = width
        self.__height = height
        self.__vertical = vertical
        self.__normalize_tex_coord = normalize_tex_coord
        self.start_building()

    def build(self):
        self.is_closed = False
        self.self_calculated_normal = True

        vertices = self.vertices
        indices = self.indices        
        width = self.__width
        height = self.__height
        vertical = self.__vertical
        normalize_tex_coord = self.__normalize_tex_coord

        normal = glm.vec3(0, 0, 1)
        if vertical:
            normal = glm.vec3(0, -1, 0)

        # 左下
        vertex0 = Vertex()
        vertex0.position = glm.vec3(-width/2, -height/2, 0)
        if vertical:
            vertex0.position = glm.vec3(-width/2, 0, -height/2)
        vertex0.normal = normal
        vertex0.tex_coord = glm.vec3(0, 1-height, 0)
        if normalize_tex_coord:
            vertex0.tex_coord = glm.vec3(0, 1-height/width, 0)

        # 右下
        vertex1 = Vertex()
        vertex1.position = glm.vec3(width/2, -height/2, 0)
        if vertical:
            vertex1.position = glm.vec3(width/2, 0, -height/2)
        vertex1.normal = normal
        vertex1.tex_coord = glm.vec3(width, 1-height, 0)
        if normalize_tex_coord:
            vertex1.tex_coord = glm.vec3(1, 1-height/width, 0)

        # 右上
        vertex2 = Vertex()
        vertex2.position = glm.vec3(width/2, height/2, 0)
        if vertical:
            vertex2.position = glm.vec3(width/2, 0, height/2)
        vertex2.normal = normal
        vertex2.tex_coord = glm.vec3(width, 1, 0)
        if normalize_tex_coord:
            vertex2.tex_coord = glm.vec3(1, 1, 0)

        # 左上
        vertex3 = Vertex()
        vertex3.position = glm.vec3(-width/2, height/2, 0)
        if vertical:
            vertex3.position = glm.vec3(-width/2, 0, height/2)
        vertex3.normal = normal
        vertex3.tex_coord = glm.vec3(0, 1, 0)

        vertices[0] = vertex0
        vertices[1] = vertex1
        vertices[2] = vertex2
        vertices[3] = vertex3

        indices[0] = glm.uvec3(0, 1, 2)
        indices[1] = glm.uvec3(0, 2, 3)

        self.generate_temp_TBN(vertices[0], vertices[1], vertices[2])
        self.generate_temp_TBN(vertices[0], vertices[2], vertices[3])

        del vertices[4:]
        del indices[2:]

    @property
    def vertical(self):
        return self.__vertical
    
    @vertical.setter
    @Mesh.param_setter
    def vertical(self, flag:bool):
        self.__vertical = flag

    @property
    def width(self):
        return self.__width
    
    @width.setter
    @Mesh.param_setter
    def width(self, width:float):
        self.__width = width

    @property
    def height(self):
        return self.__height
    
    @height.setter
    @Mesh.param_setter
    def height(self, height:float):
        self.__height = height

    @property
    def normalize_tex_coord(self):
        return self.__normalize_tex_coord
    
    @normalize_tex_coord.setter
    @Mesh.param_setter
    def normalize_tex_coord(self, flag:bool):
        self.__normalize_tex_coord = flag


from glass import sampler2D
from glass.DictList import DictList
from .PostProcessEffect import PostProcessEffect

from ..Frame import Frame

class PostProcessEffects(DictList):

    def __init__(self):
        DictList.__init__(self)
        self._screen_update_time = 0
        self._should_update = False

        self.camera = None
        self.view_pos_map = None
        self.view_normal_map = None
        self.depth_map = None

    @property
    def has_valid(self)->bool:
        for effect in self:
            if effect.enabled:
                return True
            
        return False
    
    @property
    def need_pos_info(self)->bool:
        for effect in self:
            if effect.enabled and effect.need_pos_info():
                return True
            
        return False

    @property
    def last_valid(self)->PostProcessEffect:
        for i in range(len(self)-1, -1, -1):
            effect = self[i]
            if effect.enabled:
                return effect
            
        return None

    def apply(self, screen_image:sampler2D)->sampler2D:
        self._should_update = False
        for effect in self:
            if not effect.enabled:
                continue

            effect.depth_map = self.depth_map
            effect.camera = self.camera
            effect.view_pos_map = self.view_pos_map
            effect.view_normal_map = self.view_normal_map
            screen_image = effect.apply(screen_image)

            self._should_update = self._should_update or effect.should_update
            effect.should_update = False

        return screen_image
    
    @property
    def should_update(self)->bool:
        return self._should_update

    def draw_to_active(self, screen_image:sampler2D)->bool:
        should_update = False
        last_effect = None
        last_effect_index = -1
        for i in range(len(self)-1, -1, -1):
            effect = self[i]
            if effect.enabled:
                last_effect = effect
                last_effect_index = i
                break

        if last_effect is None:
            Frame.draw(screen_image)
            return

        for i in range(last_effect_index):
            effect = self[i]
            if not effect.enabled:
                continue
            
            effect.depth_map = self.depth_map
            effect.camera = self.camera
            effect.view_pos_map = self.view_pos_map
            effect.view_normal_map = self.view_normal_map
            screen_image = effect.apply(screen_image)

            should_update = effect.should_update or should_update
            effect.should_update = False

        last_effect.depth_map = self.depth_map
        last_effect.camera = self.camera
        last_effect.view_pos_map = self.view_pos_map
        last_effect.view_normal_map = self.view_normal_map
        last_effect.draw_to_active(screen_image)

        should_update = last_effect.should_update or should_update
        last_effect.should_update = False

        return should_update
    
    @property
    def screen_update_time(self)->float:
        return self._screen_update_time
    
    @screen_update_time.setter
    def screen_update_time(self, screen_update_time:float)->None:
        self._screen_update_time = screen_update_time
        for effect in self:
            effect.screen_update_time = screen_update_time
    


from .WeakList import WeakList

class ExtendableList:

    def __init__(self, weak_ref:bool=True)->None:
        if weak_ref:
            self._list = WeakList()
        else:
            self._list = []

    def __setitem__(self, index:(int,slice), value:object)->None:
        len_self = len(self)
        if isinstance(index, int) and index >= len_self:
            self._list.extend([None]*(index-len_self+1))

        self._list[index] = value

    def __getitem__(self, index:(int,slice))->object:
        return self._list[index]
            
    def __delitem__(self, index:(int,slice))->None:
        del self._list[index]
        
    def __contains__(self, value:object)->bool:
        return value in self._list
        
    def __len__(self)->int:
        return self._list.__len__()
    
    def __bool__(self)->bool:
        return bool(self._list)
    
    def __iter__(self)->type(iter([])):
        return self._list.__iter__()
    
    def append(self, value):
        self._list.append(value)

    def insert(self, index, value):
        self._list.insert(index, value)

    def extend(self, value_list):
        self._list.extend(value_list)

    def remove(self, value):
        self._list.remove(value)

    def pop(self, index):
        return self._list.pop(index)
        
    def clear(self):
        self._list.clear()

class DictList(ExtendableList):
    def __init__(self, values:(list,dict)=None, weak_ref:bool=False, **kwargs):
        ExtendableList.__init__(self, weak_ref)

        self._key_index_map = {}
        self._key_list = ExtendableList()

        if isinstance(values, list):
            self.extend(values)
        elif isinstance(values, dict):
            self.update(values)

        self.update(kwargs)
    
    def __setitem__(self, index:(int,slice,str), value):
        if isinstance(index, str):
            key = index
            if key in self._key_index_map:
                index = self._key_index_map[key]
                ExtendableList.__setitem__(self, index, value)
            else:
                len_self = len(self)
                self._key_index_map[key] = len_self
                self._key_list[len_self] = key
                ExtendableList.append(self, value)
        else:
            ExtendableList.__setitem__(self, index, value)

    def __getitem__(self, index:(int,slice,str)):
        if isinstance(index, str):
            key = index
            index = self._key_index_map[key]

        return ExtendableList.__getitem__(self, index)
    
    def __update_after(self, index):
        for i in range(index, len(self._key_list)):
            key = self._key_list[i]
            if key is not None:
                self._key_index_map[key] = i

    def __delitem__(self, index:(int,str,slice)):
        if isinstance(index, slice):
            start, stop, step = self._process_slice(index)
            for i in range(start, stop, step):
                del self[i]
            return

        key = None
        if isinstance(index, str):
            key = index
            index = self._key_index_map[key]
        elif isinstance(index, int):
            key = self._key_list[index]

        if key is not None:
            del self._key_index_map[key]

        del self._key_list[index]
        ExtendableList.__delitem__(self, index)
        
        self.__update_after(index)

    def pop(self, index:(int,str)):
        key = None
        if isinstance(index, str):
            key = index
            index = self._key_index_map[key]
        elif isinstance(index, int):
            key = self._key_list[index]

        if key is not None:
            del self._key_index_map[key]

        del self._key_list[index]
        value = ExtendableList.pop(self, index)
        
        self.__update_after(index)

        return value
        
    def insert(self, index:(int,str), value, key:str=None):
        if isinstance(index, str):
            index = self._key_index_map[index]

        ExtendableList.insert(self, index, value)
        self._key_list.insert(index, key)
        if key is not None:
            self._key_index_map[key] = index

        self.__update_after(index+1)

    def append(self, value, key:str=None):
        ExtendableList.append(self, value)
        self._key_list.append(key)
        if key is not None:
            self._key_index_map[key] = len(self)-1

    def extend(self, new_list):
        ExtendableList.extend(self, new_list)
        self._key_list.extend([None]*len(new_list))
    
    def update(self, new_dict):
        for key, value in new_dict.items():
            self[key] = value

    def index(self, key:str)->int:
        return self._key_index_map[key]

    def key(self, index:int)->str:
        return self._key_list[index]
        
    def clear(self):
        ExtendableList.clear(self)
        self._key_index_map.clear()
        self._key_list.clear()
        
    def __contains__(self, key)->bool:
        if isinstance(key, str):
            return (key in self._key_index_map)
        else:
            return ExtendableList.__contains__(self, key)

    def keys(self):
        return self._key_index_map.keys()


from .SceneNode import SceneNode
from .Mesh import Mesh
from .Material import Material

from glass.download import download, is_China_user
from glass.utils import checktype
from glass.AttrList import AttrList
from glass import Vertices, sampler2D, Indices, GLInfo
from glass.ImageLoader import ImageLoader

import os
import sys
import platform

import glm
from OpenGL import GL
from enum import Flag

ModelLoader = None

def import_AssimpModelLoader()->None:
    global ModelLoader
    if ModelLoader is not None:
        return
    
    self_folder = os.path.dirname(os.path.abspath(__file__))
    versions = platform.python_version().split(".")
    postfix = "-win_amd64.pyd" if platform.architecture()[0] == "64bit" else "-win32.pyd"

    module_folder = self_folder + "/AssimpModelLoader"
    if postfix == "-win32.pyd":
        module_folder += "_win32"

    md5_map = \
    {
        "assimp-vc143-mt.dll-win_amd64.pyd": "739010c640f5cc8472e9545bbec55215",
        "AssimpModelLoader.cp36-win_amd64.pyd": "4b88c7af14add7c7a97303257cafcc39",
        "AssimpModelLoader.cp37-win_amd64.pyd": "1264d2b5c4ba736c0cfc97ea64b0c980",
        "AssimpModelLoader.cp38-win_amd64.pyd": "24f18a3c43f66af2060cf09a7f2e9fbf",
        "AssimpModelLoader.cp39-win_amd64.pyd": "0e634908142f44a5995ec761630fdad1",
        "AssimpModelLoader.cp310-win_amd64.pyd": "e4ac67a9a787f7a6dfb3ff360664f0ca",
        "AssimpModelLoader.cp311-win_amd64.pyd": "abaf2bf9a2acf2681efdb53abeeeb4af",
        "AssimpModelLoader.cp312-win_amd64.pyd": "9ff1b2b7c572d9617cf9ba0ca787f963",

        "assimp-vc143-mt.dll-win32.pyd": "5e10d5946ee44584eda250ac458da889",
        "AssimpModelLoader.cp36-win32.pyd": "32f734260138847c8734ac3dd3b1919d",
        "AssimpModelLoader.cp37-win32.pyd": "692354b624b87c4e17a265f5db1475a8",
        "AssimpModelLoader.cp38-win32.pyd": "f0ffa3502bb9b7c225e6bf20856932e6",
        "AssimpModelLoader.cp39-win32.pyd": "762aa5a6b281b894ec955724ba613f19",
        "AssimpModelLoader.cp310-win32.pyd": "563677469a4daa43ffc70a40dd993695",
        "AssimpModelLoader.cp311-win32.pyd": "ceb6d0ce38f44ad82379c5298106ddbb",
        "AssimpModelLoader.cp312-win32.pyd": "822fb1a01ca05f547033da5acacc233d",
    }

    pyd_name = "AssimpModelLoader.cp3" + versions[1] + postfix
    target_pyd = module_folder + "/" + pyd_name
    gitee_url = ""
    github_url = ""
    if postfix == "-win_amd64.pyd":
        gitee_url = "https://gitee.com/time-coder/Glass-Engine/raw/main/glass_engine/AssimpModelLoader/" + pyd_name
        github_url = "https://raw.githubusercontent.com/Time-Coder/Glass-Engine/main/glass_engine/AssimpModelLoader/" + pyd_name
    else:
        gitee_url = "https://gitee.com/time-coder/Glass-Engine/raw/main/glass_engine/AssimpModelLoader_win32/" + pyd_name
        github_url = "https://raw.githubusercontent.com/Time-Coder/Glass-Engine/main/glass_engine/AssimpModelLoader_win32/" + pyd_name
    
    if is_China_user():
        download(gitee_url, target_pyd, md5_map[pyd_name])
    else:
        download(github_url, target_pyd, md5_map[pyd_name])

    dll_name = "assimp-vc143-mt.dll"
    target_dll = module_folder + "/" + dll_name
    if postfix == "-win_amd64.pyd":
        gitee_url = "https://gitee.com/time-coder/Glass-Engine/raw/main/glass_engine/AssimpModelLoader/assimp-vc143-mt.dll"
        github_url = "https://raw.githubusercontent.com/Time-Coder/Glass-Engine/main/glass_engine/AssimpModelLoader/assimp-vc143-mt.dll"
    else:
        gitee_url = "https://gitee.com/time-coder/Glass-Engine/raw/main/glass_engine/AssimpModelLoader_win32/assimp-vc143-mt.dll"
        github_url = "https://raw.githubusercontent.com/Time-Coder/Glass-Engine/main/glass_engine/AssimpModelLoader_win32/assimp-vc143-mt.dll"

    if is_China_user():
        download(gitee_url, target_dll, md5_map[dll_name + postfix])
    else:
        download(github_url, target_dll, md5_map[dll_name + postfix])

    if module_folder not in sys.path:
        sys.path.append(module_folder)

    import AssimpModelLoader
    ModelLoader = AssimpModelLoader

class AssimpImportError(Exception):
    pass

class ModelMesh(Mesh):

    def __init__(self, assimp_mesh):
        Mesh.__init__(self, name=assimp_mesh.name, primitive_type=GLInfo.enum_map[assimp_mesh.primitive_type])
        self.__assimp_mesh = assimp_mesh
        self.start_building()

    def build(self):
        assimp_mesh = self.__assimp_mesh

        self.x_min = assimp_mesh.x_min
        self.x_max = assimp_mesh.x_max
        self.y_min = assimp_mesh.y_min
        self.y_max = assimp_mesh.y_max
        self.z_min = assimp_mesh.z_min
        self.z_max = assimp_mesh.z_max

        self.should_add_color = False
        self.vertices = Vertices(
            position=AttrList.frombuffer(assimp_mesh.position_buffer, glm.vec3),
            tangent=AttrList.frombuffer(assimp_mesh.tangent_buffer, glm.vec3),
            bitangent=AttrList.frombuffer(assimp_mesh.bitangent_buffer, glm.vec3),
            normal=AttrList.frombuffer(assimp_mesh.normal_buffer, glm.vec3),
            tex_coord=AttrList.frombuffer(assimp_mesh.tex_coord_buffers[0], glm.vec3),
            color=AttrList.frombuffer(assimp_mesh.color_buffers[0], glm.vec4),
            back_color=AttrList.frombuffer(assimp_mesh.color_buffers[1], glm.vec4)
        )
        self.indices = Indices.frombuffer(assimp_mesh.indices_buffer, glm.uvec3)

class Model(SceneNode):

    class PostProcessSteps(Flag):
        Nothing = 0x0
        CalcTangentSpace = 0x1
        JoinIdenticalVertices = 0x2
        MakeLeftHanded = 0x4
        Triangulate = 0x8
        RemoveComponent = 0x10
        GenNormals = 0x20
        GenSmoothNormals = 0x40
        SplitLargeMeshes = 0x80
        PreTransformVertices = 0x100
        LimitBoneWeights = 0x200
        ValidateDataStructure = 0x400
        ImproveCacheLocality = 0x800
        RemoveRedundantMaterials = 0x1000
        FixInfacingNormals = 0x2000
        PopulateArmatureData = 0x4000
        SortByPType = 0x8000
        FindDegenerates = 0x10000
        FindInvalidData = 0x20000
        GenUVCoords = 0x40000
        TransformUVCoords = 0x80000
        FindInstances = 0x100000
        OptimizeMeshes  = 0x200000
        OptimizeGraph  = 0x400000
        FlipUVs = 0x800000
        FlipWindingOrder  = 0x1000000
        SplitByBoneCount  = 0x2000000
        Debone  = 0x4000000
        GlobalScale = 0x8000000
        EmbedTextures  = 0x10000000
        ForceGenNormals = 0x20000000
        DropNormals = 0x40000000
        GenBoundingBoxes = 0x80000000

        default = (SortByPType |
                   ValidateDataStructure |
                   SplitLargeMeshes |
                   JoinIdenticalVertices |
                   Triangulate |
                   CalcTangentSpace |
                   GenNormals |
                   GenBoundingBoxes)

    @checktype
    def __init__(self, file_name:str="", flags:PostProcessSteps=PostProcessSteps.default, extra_flags:PostProcessSteps=PostProcessSteps.Nothing, exclude_flags:PostProcessSteps=PostProcessSteps.Nothing):
        self.__flags = ((flags | extra_flags) & (~exclude_flags))

        if file_name:
            name = os.path.basename(file_name)
            SceneNode.__init__(self, name)
            self.load(file_name, flags=self.__flags)
        else:
            SceneNode.__init__(self, "")
    
    @property
    def flags(self):
        return self.__flags

    @checktype
    def load(self, file_name:str, flags:PostProcessSteps=PostProcessSteps.default, extra_flags:PostProcessSteps=PostProcessSteps.Nothing, exclude_flags:PostProcessSteps=PostProcessSteps.Nothing):
        if not os.path.isfile(file_name):
            raise FileNotFoundError(file_name)
        
        self.__flags = ((flags | extra_flags) & (~exclude_flags))
        self.__file_name = os.path.abspath(file_name)
        self.__dir_name = os.path.dirname(file_name)

        self["root"].clear_children()
        self.__materials = []
        self.__meshes = []

        import_AssimpModelLoader()
        assimp_model = ModelLoader.load(self.__file_name, flags.value)

        if not assimp_model.success:
            raise AssimpImportError(assimp_model.error_message)
        if not assimp_model.nodes or \
           (not assimp_model.nodes[0].children and not assimp_model.nodes[0].meshes):
            return

        for assimp_material in assimp_model.materials:
            material = Material(name=assimp_material.name)

            ambient = glm.vec3(assimp_material.ambient.r, assimp_material.ambient.g, assimp_material.ambient.b)
            diffuse = glm.vec3(assimp_material.diffuse.r, assimp_material.diffuse.g, assimp_material.diffuse.b)
            specular = glm.vec3(assimp_material.specular.r, assimp_material.specular.g, assimp_material.specular.b)
            emission = glm.vec3(assimp_material.emission.r, assimp_material.emission.g, assimp_material.emission.b)
            reflection = glm.vec4(assimp_material.reflection.r, assimp_material.reflection.g, assimp_material.reflection.b, assimp_material.reflection.a)
            base_color = glm.vec3(assimp_material.base_color.r, assimp_material.base_color.g, assimp_material.base_color.b)
            shininess = assimp_material.shininess
            shininess_strength = assimp_material.shininess_strength
            opacity = assimp_material.opacity
            refractive_index = assimp_material.refractive_index
            roughness = assimp_material.roughness
            metallic = assimp_material.metallic

            if glm.length(ambient) > 0:
                material.ambient = ambient

            if glm.length(diffuse) > 0:
                material.diffuse = diffuse

            if glm.length(specular) > 0:
                material.specular = specular

            if glm.length(emission) > 0:
                material.emission = emission

            if glm.length(reflection) > 0:
                material.reflection = reflection

            if glm.length(base_color) > 0:
                material.base_color = base_color

            if refractive_index > 0:
                material.refractive_index = refractive_index

            if shininess > 0:
                material.shininess = shininess

            if shininess_strength > 0:
                material.shininess_strength = shininess_strength

            material.shading_model = Material.ShadingModel(assimp_material.shading_model.value)
            material.roughness = roughness
            material.metallic = metallic
            material.opacity = opacity
            texture_maps = \
            [
                "ambient_map", "diffuse_map", "specular_map",
                "shininess_map", "emission_map", "height_map",
                "normal_map", "opacity_map", "reflection_map",
                "base_color_map", "ao_map", "roughness_map", "metallic_map"
            ]

            arm_map = None
            for texture_map in texture_maps:
                texture_map_list = getattr(assimp_material, texture_map)
                if texture_map_list:
                    image_path = self.__dir_name + "/" + texture_map_list[0]
                    
                    if texture_map == "roughness_map" and image_path == arm_map:
                        image = ImageLoader.load(image_path)
                        if len(image.shape) > 2 and image.shape[2] >= 3:
                            material.arm_map = sampler2D.load(image_path)
                            material.ao_map = None
                            break

                    if texture_map == "metallic_map" and image_path == arm_map:
                        image = ImageLoader.load(image_path)
                        if len(image.shape) > 2 and image.shape[2] >= 3:
                            material.arm_map = sampler2D.load(image_path)
                            material.ao_map = None
                            material.roughness_map = None
                            break

                    if texture_map in ["ao_map", "roughness_map", "metallic_map"]:
                        arm_map = image_path

                    if os.path.isfile(image_path):
                        if texture_map == "ambient_map":
                            image = ImageLoader.load(image_path)
                            image_dtype = image.dtype
                            threshold = 0.5
                            if "int" in str(image_dtype):
                                threshold = 127

                            if image.max() > threshold:
                                image = (0.2*image).astype(image_dtype)

                            material.ambient_map = sampler2D(image)
                        else:
                            setattr(material, texture_map, sampler2D.load(image_path))

            self.__materials.append(material)

        len_materials = len(self.__materials)
        for assimp_mesh in assimp_model.meshes:
            mesh = ModelMesh(assimp_mesh=assimp_mesh)
            material_index = assimp_mesh.material_index
            if 0 <= material_index < len_materials:
                mesh.material = self.__materials[material_index]

                assimp_material = assimp_model.materials[material_index]
                if assimp_material.wireframe:
                    mesh.render_hints.polygon_mode = GL.GL_LINE
                if assimp_material.twoside:
                    mesh.render_hints.cull_face = None
            
            self.__meshes.append(mesh)

        self.__load_node(assimp_model.nodes[0], assimp_model, self["root"])
    
    def __load_node(self, assimp_node, assimp_model, parent_node=None):
        node = parent_node
        if node is None:
            node = SceneNode(name=assimp_node.name)

        node.orientation.w = assimp_node.orientation.w
        node.orientation.x = assimp_node.orientation.x
        node.orientation.y = assimp_node.orientation.y
        node.orientation.z = assimp_node.orientation.z

        node.position.x = assimp_node.position.x
        node.position.y = assimp_node.position.y
        node.position.z = assimp_node.position.z

        node.scale.x = assimp_node.scale.x
        node.scale.y = assimp_node.scale.y
        node.scale.z = assimp_node.scale.z

        for mesh_index in assimp_node.meshes:
            node.add_child(self.__meshes[mesh_index])
        for node_index in assimp_node.children:
            assimp_child = assimp_model.nodes[node_index]
            if not assimp_child.meshes and not assimp_child.children:
                continue
            node.add_child(self.__load_node(assimp_model.nodes[node_index], assimp_model))
        return node



from glass.download import pip_install

try:
    import PyQt6
except:
    pip_install("PyQt6")

from PyQt6.QtWidgets import QSplitter, QDialog, QCheckBox, QWidget, QVBoxLayout, QApplication, QHBoxLayout, QScrollArea
from PyQt6.QtGui import QIcon
from PyQt6.QtCore import pyqtSignal, Qt
from qt_material import apply_stylesheet

from .QtComponents import IntInput, ComboboxInput, MapInput, SliderInput, ColorImageInput, ValueImageInput, ColorImageChooser

from ..Model import Model
from ..BasicScene import ModelView
from ..Geometries.Floor import Floor
from ..Geometries.Sphere import Sphere
from ..Material import Material

import sys
import os
import functools
import glm

def slot(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        self = args[0]
        if not self.should_call_slot:
            return
        return func(*args, **kwargs)
    return wrapper

class MaterialEditor(QWidget):

    material_changed = pyqtSignal()
    internal_changed = pyqtSignal()

    def __init__(self, material:Material, parent:QWidget=None):
        QWidget.__init__(self, parent=parent)

        self.__material = None
        self.inputs = {}
        self.should_call_slot = True
        self.parameters_map = \
        {
            "ambient": {Material.ShadingModel.Flat, Material.ShadingModel.Gouraud, Material.ShadingModel.Phong, Material.ShadingModel.PhongBlinn, Material.ShadingModel.Toon, Material.ShadingModel.OrenNayar, Material.ShadingModel.Minnaert, Material.ShadingModel.CookTorrance, Material.ShadingModel.Fresnel, Material.ShadingModel.PBR},
            "diffuse": {Material.ShadingModel.Flat, Material.ShadingModel.Gouraud, Material.ShadingModel.Phong, Material.ShadingModel.PhongBlinn, Material.ShadingModel.Toon, Material.ShadingModel.OrenNayar, Material.ShadingModel.Minnaert, Material.ShadingModel.Fresnel, Material.ShadingModel.Unlit},
            "specular": {Material.ShadingModel.Gouraud, Material.ShadingModel.Phong, Material.ShadingModel.PhongBlinn, Material.ShadingModel.Toon},
            "shininess_strength": {Material.ShadingModel.Gouraud, Material.ShadingModel.Phong, Material.ShadingModel.PhongBlinn, Material.ShadingModel.Toon},
            "shininess": {Material.ShadingModel.Gouraud, Material.ShadingModel.Phong, Material.ShadingModel.PhongBlinn, Material.ShadingModel.Toon},
            "emission": {Material.ShadingModel.Flat, Material.ShadingModel.Gouraud, Material.ShadingModel.Phong, Material.ShadingModel.PhongBlinn, Material.ShadingModel.Toon, Material.ShadingModel.OrenNayar, Material.ShadingModel.Minnaert, Material.ShadingModel.CookTorrance, Material.ShadingModel.Fresnel, Material.ShadingModel.PBR, Material.ShadingModel.Unlit},
            "emission_strength": {Material.ShadingModel.Flat, Material.ShadingModel.Gouraud, Material.ShadingModel.Phong, Material.ShadingModel.PhongBlinn, Material.ShadingModel.Toon, Material.ShadingModel.OrenNayar, Material.ShadingModel.Minnaert, Material.ShadingModel.CookTorrance, Material.ShadingModel.Fresnel, Material.ShadingModel.PBR, Material.ShadingModel.Unlit},
            "rim_power": {Material.ShadingModel.Flat, Material.ShadingModel.Gouraud, Material.ShadingModel.Phong, Material.ShadingModel.PhongBlinn, Material.ShadingModel.Toon, Material.ShadingModel.OrenNayar, Material.ShadingModel.Minnaert, Material.ShadingModel.CookTorrance, Material.ShadingModel.Fresnel, Material.ShadingModel.PBR},
            "base_color": {Material.ShadingModel.CookTorrance, Material.ShadingModel.PBR},
            "metallic": {Material.ShadingModel.CookTorrance, Material.ShadingModel.PBR},
            "roughness": {Material.ShadingModel.CookTorrance, Material.ShadingModel.PBR, Material.ShadingModel.OrenNayar, Material.ShadingModel.Minnaert},
            "diffuse_bands": {Material.ShadingModel.Toon},
            "specular_bands": {Material.ShadingModel.Toon},
            "diffuse_softness": {Material.ShadingModel.Toon},
            "specular_softness": {Material.ShadingModel.Toon},
            "reflection": {Material.ShadingModel.Flat, Material.ShadingModel.Gouraud, Material.ShadingModel.Phong, Material.ShadingModel.PhongBlinn, Material.ShadingModel.Toon, Material.ShadingModel.OrenNayar, Material.ShadingModel.Minnaert, Material.ShadingModel.CookTorrance, Material.ShadingModel.Fresnel, Material.ShadingModel.PBR},
            "reflection_opacity": {Material.ShadingModel.Flat, Material.ShadingModel.Gouraud, Material.ShadingModel.Phong, Material.ShadingModel.PhongBlinn, Material.ShadingModel.Toon, Material.ShadingModel.OrenNayar, Material.ShadingModel.Minnaert, Material.ShadingModel.CookTorrance, Material.ShadingModel.Fresnel, Material.ShadingModel.PBR},
            "refractive_index": {Material.ShadingModel.Flat, Material.ShadingModel.Gouraud, Material.ShadingModel.Phong, Material.ShadingModel.PhongBlinn, Material.ShadingModel.Toon, Material.ShadingModel.OrenNayar, Material.ShadingModel.Minnaert, Material.ShadingModel.CookTorrance, Material.ShadingModel.Fresnel, Material.ShadingModel.PBR},
            "dynamic_env_mapping": {Material.ShadingModel.Flat, Material.ShadingModel.Gouraud, Material.ShadingModel.Phong, Material.ShadingModel.PhongBlinn, Material.ShadingModel.Toon, Material.ShadingModel.OrenNayar, Material.ShadingModel.Minnaert, Material.ShadingModel.CookTorrance, Material.ShadingModel.Fresnel, Material.ShadingModel.PBR},
            "env_mix_diffuse": {Material.ShadingModel.Flat, Material.ShadingModel.Gouraud, Material.ShadingModel.Phong, Material.ShadingModel.PhongBlinn, Material.ShadingModel.Toon, Material.ShadingModel.OrenNayar, Material.ShadingModel.Minnaert, Material.ShadingModel.CookTorrance, Material.ShadingModel.Fresnel, Material.ShadingModel.PBR},
            "opacity": {Material.ShadingModel.Flat, Material.ShadingModel.Gouraud, Material.ShadingModel.Phong, Material.ShadingModel.PhongBlinn, Material.ShadingModel.Toon, Material.ShadingModel.OrenNayar, Material.ShadingModel.Minnaert, Material.ShadingModel.CookTorrance, Material.ShadingModel.Fresnel, Material.ShadingModel.PBR},
            "height_scale": {Material.ShadingModel.Flat, Material.ShadingModel.Gouraud, Material.ShadingModel.Phong, Material.ShadingModel.PhongBlinn, Material.ShadingModel.Toon, Material.ShadingModel.OrenNayar, Material.ShadingModel.Minnaert, Material.ShadingModel.CookTorrance, Material.ShadingModel.Fresnel, Material.ShadingModel.PBR},
            "fog": {Material.ShadingModel.Flat, Material.ShadingModel.Gouraud, Material.ShadingModel.Phong, Material.ShadingModel.PhongBlinn, Material.ShadingModel.Toon, Material.ShadingModel.OrenNayar, Material.ShadingModel.Minnaert, Material.ShadingModel.CookTorrance, Material.ShadingModel.Fresnel, Material.ShadingModel.PBR},
            "recv_shadows": {Material.ShadingModel.Flat, Material.ShadingModel.Gouraud, Material.ShadingModel.Phong, Material.ShadingModel.PhongBlinn, Material.ShadingModel.Toon, Material.ShadingModel.OrenNayar, Material.ShadingModel.Minnaert, Material.ShadingModel.CookTorrance, Material.ShadingModel.Fresnel, Material.ShadingModel.PBR},
            "cast_shadows": {Material.ShadingModel.Flat, Material.ShadingModel.Gouraud, Material.ShadingModel.Phong, Material.ShadingModel.PhongBlinn, Material.ShadingModel.Toon, Material.ShadingModel.OrenNayar, Material.ShadingModel.Minnaert, Material.ShadingModel.CookTorrance, Material.ShadingModel.Fresnel, Material.ShadingModel.PBR},
            "ao_map": {Material.ShadingModel.Flat, Material.ShadingModel.Gouraud, Material.ShadingModel.Phong, Material.ShadingModel.PhongBlinn, Material.ShadingModel.Toon, Material.ShadingModel.OrenNayar, Material.ShadingModel.Minnaert, Material.ShadingModel.CookTorrance, Material.ShadingModel.Fresnel, Material.ShadingModel.PBR},
            "normal_map": {Material.ShadingModel.Flat, Material.ShadingModel.Gouraud, Material.ShadingModel.Phong, Material.ShadingModel.PhongBlinn, Material.ShadingModel.Toon, Material.ShadingModel.OrenNayar, Material.ShadingModel.Minnaert, Material.ShadingModel.CookTorrance, Material.ShadingModel.Fresnel, Material.ShadingModel.PBR},
            "height_map": {Material.ShadingModel.Flat, Material.ShadingModel.Gouraud, Material.ShadingModel.Phong, Material.ShadingModel.PhongBlinn, Material.ShadingModel.Toon, Material.ShadingModel.OrenNayar, Material.ShadingModel.Minnaert, Material.ShadingModel.CookTorrance, Material.ShadingModel.Fresnel, Material.ShadingModel.PBR},
            "arm_map": {Material.ShadingModel.Flat, Material.ShadingModel.Gouraud, Material.ShadingModel.Phong, Material.ShadingModel.PhongBlinn, Material.ShadingModel.Toon, Material.ShadingModel.OrenNayar, Material.ShadingModel.Minnaert, Material.ShadingModel.CookTorrance, Material.ShadingModel.Fresnel, Material.ShadingModel.PBR},
        }

        vlayout_container = QVBoxLayout()
        scroll_area = QScrollArea()

        main_widget = QWidget()
        vlayout = QVBoxLayout()
        shading_model_values = \
        {
            "Flat": Material.ShadingModel.Flat,
            "Gouraud": Material.ShadingModel.Gouraud,
            "Phong": Material.ShadingModel.Phong,
            "PhongBlinn": Material.ShadingModel.PhongBlinn,
            "Toon": Material.ShadingModel.Toon,
            "OrenNaya": Material.ShadingModel.OrenNayar,
            "Minnaert": Material.ShadingModel.Minnaert,
            "Unlit": Material.ShadingModel.Unlit,
            "Fresnel": Material.ShadingModel.Fresnel,
            "PBR": Material.ShadingModel.PBR
        }
        self.inputs["shading_model"] = ComboboxInput("着色模型", shading_model_values, parent=self)
        vlayout.addWidget(self.inputs["shading_model"])

        internal_materials_values = \
        {
            "自定义": None,
            "翡翠": Material.Type.Emerald,
            "玉":  Material.Type.Jade,
            "黑曜石": Material.Type.Obsidian,
            "珍珠": Material.Type.Pearl,
            "红宝石": Material.Type.Ruby,
            "绿松石": Material.Type.Turquoise,
            "黄铜": Material.Type.Brass,
            "青铜": Material.Type.Bronze,
            "铬": Material.Type.Chrome,
            "铜": Material.Type.Copper,
            "金": Material.Type.Gold,
            "银": Material.Type.Silver,
            "黑色塑料": Material.Type.BlackPlastic,
            "青色塑料": Material.Type.CyanPlastic,
            "绿色塑料": Material.Type.GreenPlastic,
            "红色塑料": Material.Type.RedPlastic,
            "白色塑料": Material.Type.WhitePlastic,
            "黄色塑料": Material.Type.YellowPlastic,
            "黑色橡胶": Material.Type.BlackRubber,
            "青色橡胶": Material.Type.CyanRubber,
            "绿色橡胶": Material.Type.GreenRubber,
            "红色橡胶": Material.Type.RedRubber,
            "白色橡胶": Material.Type.WhiteRubber,
            "黄色橡胶": Material.Type.YellowRubber
        }
        self.inputs["internal_material"] = ComboboxInput("内置材质", internal_materials_values, parent=self)
        vlayout.addWidget(self.inputs["internal_material"])

        self.inputs["ambient"] = ColorImageInput(prompt="环境光", parent=self)
        vlayout.addWidget(self.inputs["ambient"])

        self.inputs["diffuse"] = ColorImageInput(prompt="漫反射", parent=self)
        vlayout.addWidget(self.inputs["diffuse"])

        self.inputs["specular"] = ColorImageInput(prompt="镜面高光", parent=self)
        vlayout.addWidget(self.inputs["specular"])

        self.inputs["shininess_strength"] = SliderInput(prompt="高光强度", range=(1, 10), parent=self)
        vlayout.addWidget(self.inputs["shininess_strength"])

        self.inputs["shininess"] = SliderInput(prompt="闪耀度", range=(0, 256), parent=self)
        vlayout.addWidget(self.inputs["shininess"])

        self.inputs["emission"] = ColorImageInput(prompt="自发光", parent=self)
        vlayout.addWidget(self.inputs["emission"])

        self.inputs["emission_strength"] = SliderInput(prompt="自发光强度", range=(1, 50), parent=self)
        vlayout.addWidget(self.inputs["emission_strength"])

        self.inputs["rim_power"] = SliderInput(prompt="边缘光强度", range=(0, 1), parent=self)
        vlayout.addWidget(self.inputs["rim_power"])

        self.inputs["base_color"] = ColorImageInput(prompt="基础颜色", parent=self)
        vlayout.addWidget(self.inputs["base_color"])

        self.inputs["metallic"] = ValueImageInput(prompt="金属度", parent=self)
        vlayout.addWidget(self.inputs["metallic"])

        self.inputs["roughness"] = ValueImageInput(prompt="粗糙度", parent=self)
        vlayout.addWidget(self.inputs["roughness"])

        self.inputs["diffuse_bands"] = IntInput(prompt="漫反射条带数", range=(2, 5), parent=self)
        vlayout.addWidget(self.inputs["diffuse_bands"])

        self.inputs["specular_bands"] = IntInput(prompt="镜面高光条带数", range=(2, 5), parent=self)
        vlayout.addWidget(self.inputs["specular_bands"])

        self.inputs["diffuse_softness"] = SliderInput(prompt="漫反射过度软度", range=(0, 1), parent=self)
        vlayout.addWidget(self.inputs["diffuse_softness"])

        self.inputs["specular_softness"] = SliderInput(prompt="镜面高光过度软度", range=(0, 1), parent=self)
        vlayout.addWidget(self.inputs["specular_softness"])

        self.inputs["opacity"] = ValueImageInput(prompt="不透明度", parent=self)
        vlayout.addWidget(self.inputs["opacity"])

        self.inputs["height_scale"] = SliderInput(prompt="凹凸夸张系数", range=(0, 1), parent=self)
        vlayout.addWidget(self.inputs["height_scale"])

        self.inputs["reflection"] = ColorImageInput(prompt="环境映射层", parent=self)
        vlayout.addWidget(self.inputs["reflection"])

        self.inputs["reflection_opacity"] = SliderInput(prompt="环境映射层不透明度", range=(0,1), parent=self)
        vlayout.addWidget(self.inputs["reflection_opacity"])

        self.inputs["refractive_index"] = SliderInput(prompt="折射率", range=(1,2), checkable=True, parent=self)
        vlayout.addWidget(self.inputs["refractive_index"])

        self.inputs["dynamic_env_mapping"] = QCheckBox("动态环境映射", parent=self)
        vlayout.addWidget(self.inputs["dynamic_env_mapping"])

        self.inputs["env_mix_diffuse"] = QCheckBox("环境映射与漫反射混合", parent=self)
        vlayout.addWidget(self.inputs["env_mix_diffuse"])

        self.inputs["fog"] = QCheckBox("受雾影响", parent=self)
        vlayout.addWidget(self.inputs["fog"])

        self.inputs["recv_shadows"] = QCheckBox("接收阴影", parent=self)
        vlayout.addWidget(self.inputs["recv_shadows"])

        self.inputs["cast_shadows"] = QCheckBox("投射阴影", parent=self)
        vlayout.addWidget(self.inputs["cast_shadows"])

        self.inputs["ao_map"] = MapInput(prompt="ao 贴图", width=35, height=35, parent=self)
        vlayout.addWidget(self.inputs["ao_map"])

        self.inputs["normal_map"] = MapInput(prompt="法向量贴图", width=35, height=35, parent=self)
        vlayout.addWidget(self.inputs["normal_map"])

        self.inputs["height_map"] = MapInput(prompt="凹凸贴图", width=35, height=35, parent=self)
        vlayout.addWidget(self.inputs["height_map"])

        self.inputs["arm_map"] = MapInput(prompt="arm 贴图", width=35, height=35, parent=self)
        vlayout.addWidget(self.inputs["arm_map"])
        vlayout.addStretch()

        main_widget.setLayout(vlayout)
        scroll_area.setWidget(main_widget)

        vlayout_container.addWidget(scroll_area)
        self.setLayout(vlayout_container)

        self.material = material

        self.inputs["shading_model"].value_changed.connect(self.slot_shading_model_changed)
        self.inputs["internal_material"].value_changed.connect(self.slot_internal_material_changed)
        self.inputs["ambient"].type_changed.connect(self.slot_ambient_type_changed)
        self.inputs["ambient"].glm_color_changed.connect(self.slot_ambient_color_changed)
        self.inputs["ambient"].file_path_changed.connect(self.slot_ambient_map_changed)
        self.inputs["diffuse"].type_changed.connect(self.slot_diffuse_type_changed)
        self.inputs["diffuse"].glm_color_changed.connect(self.slot_diffuse_color_changed)
        self.inputs["diffuse"].file_path_changed.connect(self.slot_diffuse_map_changed)
        self.inputs["specular"].type_changed.connect(self.slot_specular_type_changed)
        self.inputs["specular"].glm_color_changed.connect(self.slot_specular_color_changed)
        self.inputs["specular"].file_path_changed.connect(self.slot_specular_map_changed)
        self.inputs["shininess_strength"].value_changed.connect(self.slot_shininess_strength_changed)
        self.inputs["shininess"].value_changed.connect(self.slot_shininess_changed)
        self.inputs["emission"].type_changed.connect(self.slot_emission_type_changed)
        self.inputs["emission"].glm_color_changed.connect(self.slot_emission_color_changed)
        self.inputs["emission"].file_path_changed.connect(self.slot_emission_map_changed)
        self.inputs["emission_strength"].value_changed.connect(self.slot_emission_strength_changed)
        self.inputs["rim_power"].value_changed.connect(self.slot_rim_power_changed)
        self.inputs["base_color"].type_changed.connect(self.slot_base_color_type_changed)
        self.inputs["base_color"].glm_color_changed.connect(self.slot_base_color_changed)
        self.inputs["base_color"].file_path_changed.connect(self.slot_base_color_map_changed)
        self.inputs["metallic"].type_changed.connect(self.slot_metallic_type_changed)
        self.inputs["metallic"].value_changed.connect(self.slot_metallic_value_changed)
        self.inputs["metallic"].file_path_changed.connect(self.slot_metallic_map_changed)
        self.inputs["roughness"].type_changed.connect(self.slot_roughness_type_changed)
        self.inputs["roughness"].value_changed.connect(self.slot_roughness_value_changed)
        self.inputs["roughness"].file_path_changed.connect(self.slot_roughness_map_changed)
        self.inputs["diffuse_bands"].value_changed.connect(self.slot_diffuse_bands_changed)
        self.inputs["specular_bands"].value_changed.connect(self.slot_specular_bands_changed)
        self.inputs["diffuse_softness"].value_changed.connect(self.slot_diffuse_softness_changed)
        self.inputs["specular_softness"].value_changed.connect(self.slot_specular_softness_changed)
        self.inputs["reflection"].type_changed.connect(self.slot_reflection_type_changed)
        self.inputs["reflection"].glm_color_changed.connect(self.slot_reflection_color_changed)
        self.inputs["reflection"].file_path_changed.connect(self.slot_reflection_map_changed)
        self.inputs["reflection_opacity"].value_changed.connect(self.slot_reflection_opacity_changed)
        self.inputs["refractive_index"].checked_changed.connect(self.slot_refractive_index_checked_changed)
        self.inputs["refractive_index"].value_changed.connect(self.slot_refractive_index_changed)
        self.inputs["dynamic_env_mapping"].stateChanged.connect(self.slot_dynamic_env_map_changed)
        self.inputs["env_mix_diffuse"].stateChanged.connect(self.slot_env_mix_diffuse_changed)
        self.inputs["opacity"].type_changed.connect(self.slot_opacity_type_changed)
        self.inputs["opacity"].value_changed.connect(self.slot_opacity_value_changed)
        self.inputs["opacity"].file_path_changed.connect(self.slot_opacity_map_changed)
        self.inputs["height_scale"].value_changed.connect(self.slot_height_scale_changed)
        self.inputs["fog"].stateChanged.connect(self.slot_fog_changed)
        self.inputs["recv_shadows"].stateChanged.connect(self.slot_recv_shadows_changed)
        self.inputs["cast_shadows"].stateChanged.connect(self.slot_cast_shadows_changed)
        self.inputs["ao_map"].checked_changed.connect(self.slot_ao_map_checked_changed)
        self.inputs["ao_map"].file_path_changed.connect(self.slot_ao_map_changed)
        self.inputs["normal_map"].checked_changed.connect(self.slot_normal_map_checked_changed)
        self.inputs["normal_map"].file_path_changed.connect(self.slot_normal_map_changed)
        self.inputs["height_map"].checked_changed.connect(self.slot_height_map_checked_changed)
        self.inputs["height_map"].file_path_changed.connect(self.slot_height_map_changed)
        self.inputs["arm_map"].checked_changed.connect(self.slot_arm_map_checked_changed)
        self.inputs["arm_map"].file_path_changed.connect(self.slot_arm_map_changed)
        self.internal_changed.connect(self.slot_internal_changed)
        self.hide_all()

    @property
    def material(self):
        return self.__material
    
    @material.setter
    def material(self, material:Material):
        if self.__material is material:
            return
        
        self.__material = material
        self.should_call_slot = False
        self.inputs["shading_model"].value = material.shading_model
        self.inputs["ambient"].type = (ColorImageChooser.Type.Color if material.ambient_map is None else ColorImageChooser.Type.Image)
        self.inputs["ambient"].glm_color = material.ambient
        self.inputs["ambient"].file_path = ("" if material.ambient_map is None else material.ambient_map.file_name)
        self.inputs["diffuse"].type = (ColorImageChooser.Type.Color if material.diffuse_map is None else ColorImageChooser.Type.Image)
        self.inputs["diffuse"].glm_color = material.diffuse
        self.inputs["diffuse"].file_path = ("" if material.diffuse_map is None else material.diffuse_map.file_name)
        self.inputs["specular"].type = (ColorImageChooser.Type.Color if material.specular_map is None else ColorImageChooser.Type.Image)
        self.inputs["specular"].glm_color = material.specular
        self.inputs["specular"].file_path = ("" if material.specular_map is None else material.specular_map.file_name)
        self.inputs["shininess_strength"].value = material.shininess_strength
        self.inputs["shininess"].value = material.shininess
        self.inputs["emission"].type = (ColorImageChooser.Type.Color if material.emission_map is None else ColorImageChooser.Type.Image)
        self.inputs["emission"].glm_color = material.emission
        self.inputs["emission"].file_path = ("" if material.emission_map is None else material.emission_map.file_name)
        self.inputs["emission_strength"].value = material.emission_strength
        self.inputs["rim_power"].value = material.rim_power
        self.inputs["base_color"].type = (ColorImageChooser.Type.Color if material.base_color_map is None else ColorImageChooser.Type.Image)
        self.inputs["base_color"].glm_color = material.base_color
        self.inputs["base_color"].file_path = ("" if material.base_color_map is None else material.base_color_map.file_name)
        self.inputs["metallic"].type = (ValueImageInput.Type.Value if material.metallic_map is None else ValueImageInput.Type.Image)
        self.inputs["metallic"].value = material.metallic
        self.inputs["metallic"].file_path = ("" if material.metallic_map is None else material.metallic_map.file_name)
        self.inputs["roughness"].type = (ValueImageInput.Type.Value if material.roughness_map is None else ValueImageInput.Type.Image)
        self.inputs["roughness"].value = material.roughness
        self.inputs["roughness"].file_path = ("" if material.roughness_map is None else material.roughness_map.file_name)
        self.inputs["diffuse_bands"].value = material.diffuse_bands
        self.inputs["specular_bands"].value = material.specular_bands
        self.inputs["diffuse_softness"].value = material.diffuse_softness
        self.inputs["specular_softness"].value = material.specular_softness
        self.inputs["reflection"].type = (ColorImageChooser.Type.Color if material.reflection_map is None else ColorImageChooser.Type.Image)
        self.inputs["reflection"].glm_color = material.reflection.rgb
        self.inputs["reflection"].file_path = ("" if material.reflection_map is None else material.reflection_map.file_name)
        self.inputs["reflection_opacity"].value = material.reflection.a
        self.inputs["refractive_index"].value = (material.refractive_index if material.refractive_index > 1E-6 else 1.5)
        self.inputs["refractive_index"].checked = (material.refractive_index > 1E-6)
        self.inputs["dynamic_env_mapping"].setChecked(material.dynamic_env_mapping)
        self.inputs["env_mix_diffuse"].setChecked(material.env_mix_diffuse)
        self.inputs["opacity"].type = (ValueImageInput.Type.Value if material.opacity_map is None else ValueImageInput.Type.Image)
        self.inputs["opacity"].value = material.opacity
        self.inputs["opacity"].file_path = ("" if material.opacity_map is None else material.opacity_map.file_name)
        self.inputs["height_scale"].value = material.height_scale
        self.inputs["fog"].setChecked(material.fog)
        self.inputs["recv_shadows"].setChecked(material.recv_shadows)
        self.inputs["cast_shadows"].setChecked(material.cast_shadows)
        self.inputs["ao_map"].checked = (material.ao_map is not None)
        self.inputs["ao_map"].file_path = ("" if material.ao_map is None else material.ao_map.file_name)
        self.inputs["normal_map"].checked = (material.normal_map is not None)
        self.inputs["normal_map"].file_path = ("" if material.normal_map is None else material.normal_map.file_name)
        self.inputs["height_map"].checked = (material.height_map is not None)
        self.inputs["height_map"].file_path = ("" if material.height_map is None else material.height_map.file_name)
        self.inputs["arm_map"].checked = (material.arm_map is not None)
        self.inputs["arm_map"].file_path = ("" if material.arm_map is None else material.arm_map.file_name)
        self.should_call_slot = True

    def hide_all(self):
        shading_model = self.inputs["shading_model"].value
        for key, component in self.inputs.items():
            if key not in self.parameters_map:
                continue

            if shading_model in self.parameters_map[key]:
                component.show()
            else:
                component.hide()

    @slot
    def slot_internal_changed(self):
        self.should_call_slot = False
        self.inputs["internal_material"].value = None
        self.should_call_slot = True

    @slot
    def slot_shading_model_changed(self, shading_model):
        self.__material.shading_model = shading_model
        self.hide_all()
        self.material_changed.emit()

    @slot
    def slot_internal_material_changed(self, internal_material):
        self.__material.set_as(internal_material)
        self.material_changed.emit()

    @slot
    def slot_ambient_type_changed(self, type):
        if type == ColorImageChooser.Type.Color:
            self.__material.ambient_map = None
            self.__material.ambient = self.inputs["ambient"].glm_color
        elif os.path.isfile(self.inputs["ambient"].file_path):
            self.__material.ambient_map = self.inputs["ambient"].file_path
        self.material_changed.emit()
        self.internal_changed.emit()

    @slot
    def slot_ambient_color_changed(self, color):
        self.__material.ambient = color
        self.material_changed.emit()
        self.internal_changed.emit()

    @slot
    def slot_ambient_map_changed(self, ambient_map):
        self.__material.ambient_map = ambient_map
        self.material_changed.emit()
        self.internal_changed.emit()

    @slot
    def slot_diffuse_type_changed(self, type):
        if type == ColorImageChooser.Type.Color:
            self.__material.diffuse_map = None
            self.__material.diffuse = self.inputs["diffuse"].glm_color
        elif os.path.isfile(self.inputs["diffuse"].file_path):
            self.__material.diffuse_map = self.inputs["diffuse"].file_path
        self.material_changed.emit()
        self.internal_changed.emit()

    @slot
    def slot_diffuse_color_changed(self, color):
        self.__material.diffuse = color
        self.material_changed.emit()
        self.internal_changed.emit()

    @slot
    def slot_diffuse_map_changed(self, diffuse_map):
        self.__material.diffuse_map = diffuse_map
        self.material_changed.emit()
        self.internal_changed.emit()

    @slot
    def slot_specular_type_changed(self, type):
        if type == ColorImageChooser.Type.Color:
            self.__material.specular_map = None
            self.__material.specular = self.inputs["specular"].glm_color
        elif os.path.isfile(self.inputs["specular"].file_path):
            self.__material.specular_map = self.inputs["specular"].file_path
        self.material_changed.emit()
        self.internal_changed.emit()

    @slot
    def slot_specular_color_changed(self, color):
        self.__material.specular = color
        self.material_changed.emit()
        self.internal_changed.emit()

    @slot
    def slot_specular_map_changed(self, specular_map):
        self.__material.specular_map = specular_map
        self.material_changed.emit()
        self.internal_changed.emit()

    @slot
    def slot_shininess_strength_changed(self, shininess_strength):
        self.__material.shininess_strength = shininess_strength
        self.material_changed.emit()

    @slot
    def slot_shininess_changed(self, shininess):
        self.__material.shininess = shininess
        self.material_changed.emit()

    @slot
    def slot_emission_type_changed(self, type):
        if type == ColorImageChooser.Type.Color:
            self.__material.emission_map = None
            self.__material.emission = self.inputs["emission"].glm_color
        elif os.path.isfile(self.inputs["emission"].file_path):
            self.__material.emission_map = self.inputs["emission"].file_path
        self.material_changed.emit()

    @slot
    def slot_emission_color_changed(self, color):
        self.__material.emission = color
        self.material_changed.emit()

    @slot
    def slot_emission_map_changed(self, emission_map):
        self.__material.emission_map = emission_map
        self.material_changed.emit()

    @slot
    def slot_emission_strength_changed(self, emission_strength):
        self.__material.emission_strength = emission_strength
        self.material_changed.emit()

    @slot
    def slot_rim_power_changed(self, rim_power):
        self.__material.rim_power = rim_power
        self.material_changed.emit()

    @slot
    def slot_base_color_type_changed(self, type):
        if type == ColorImageChooser.Type.Color:
            self.__material.base_color_map = None
            self.__material.base_color = self.inputs["base_color"].glm_color
        elif os.path.isfile(self.inputs["base_color"].file_path):
            self.__material.base_color_map = self.inputs["base_color"].file_path
        self.material_changed.emit()

    @slot
    def slot_base_color_changed(self, color):
        self.__material.base_color = color
        self.material_changed.emit()

    @slot
    def slot_base_color_map_changed(self, base_color_map):
        self.__material.base_color_map = base_color_map
        self.material_changed.emit()

    @slot
    def slot_metallic_type_changed(self, type):
        if type == ValueImageInput.Type.Value:
            self.__material.metallic_map = None
            self.__material.metallic = self.inputs["metallic"].value
        elif os.path.isfile(self.inputs["metallic"].file_path):
            self.__material.metallic_map = self.inputs["metallic"].file_path
        self.material_changed.emit()

    @slot
    def slot_metallic_value_changed(self, metallic):
        self.__material.metallic = metallic
        self.material_changed.emit()

    @slot
    def slot_metallic_map_changed(self, metallic_map):
        self.__material.metallic_map = metallic_map
        self.material_changed.emit()

    @slot
    def slot_roughness_type_changed(self, type):
        if type == ValueImageInput.Type.Value:
            self.__material.roughness = self.inputs["roughness"].value
            self.__material.roughness_map = None
        elif os.path.isfile(self.inputs["roughness"].file_path):
            self.__material.roughness_map = self.inputs["roughness"].file_path
        self.material_changed.emit()

    @slot
    def slot_roughness_value_changed(self, value):
        self.__material.roughness = self.inputs["roughness"].value
        self.material_changed.emit()

    @slot
    def slot_roughness_map_changed(self, file_path):
        self.__material.roughness_map = file_path
        self.material_changed.emit()

    @slot
    def slot_diffuse_bands_changed(self, n_bands):
        self.__material.diffuse_bands = n_bands
        self.material_changed.emit()

    @slot
    def slot_specular_bands_changed(self, n_bands):
        self.__material.specular_bands = n_bands
        self.material_changed.emit()

    @slot
    def slot_diffuse_softness_changed(self, softness):
        self.__material.diffuse_softness = softness
        self.material_changed.emit()

    @slot
    def slot_specular_softness_changed(self, softness):
        self.__material.specular_softness = softness
        self.material_changed.emit()

    @slot
    def slot_reflection_type_changed(self, type):
        if type == ColorImageChooser.Type.Color:
            self.__material.reflection_map = None
            self.__material.reflection = glm.vec4(self.inputs["reflection"].glm_color, self.inputs["reflection_opacity"].value)
        elif os.path.isfile(self.inputs["reflection"].file_path):
            self.__material.reflection_map = self.inputs["reflection"].file_path
        
        self.material_changed.emit()

    @slot
    def slot_reflection_color_changed(self, color):
        self.__material.reflection = glm.vec4(color, self.inputs["reflection_opacity"].value)
        self.material_changed.emit()

    @slot
    def slot_reflection_map_changed(self, file_path):
        self.__material.reflection_map = file_path
        self.material_changed.emit()

    @slot
    def slot_reflection_opacity_changed(self, opacity):
        self.__material.reflection.a = opacity
        self.material_changed.emit()

    @slot
    def slot_refractive_index_changed(self, refractive_index):
        self.__material.refractive_index = refractive_index
        self.inputs["reflection"].glm_color = self.__material.reflection.rgb
        self.inputs["reflection_opacity"].value = self.__material.reflection.a
        self.material_changed.emit()

    @slot
    def slot_refractive_index_checked_changed(self, checked):
        self.__material.refractive_index = (self.inputs["refractive_index"].value if checked else 0)
        self.inputs["reflection"].glm_color = self.__material.reflection.rgb
        self.inputs["reflection_opacity"].value = self.__material.reflection.a
        self.material_changed.emit()

    @slot
    def slot_dynamic_env_map_changed(self, state):
        self.__material.dynamic_env_mapping = (state == Qt.CheckState.Checked.value)
        self.material_changed.emit()

    @slot
    def slot_env_mix_diffuse_changed(self, state):
        self.__material.env_mix_diffuse = (state == Qt.CheckState.Checked.value)
        self.material_changed.emit()

    @slot
    def slot_opacity_type_changed(self, type):
        if type == ValueImageInput.Type.Value:
            self.__material.opacity = self.inputs["opacity"].value
            self.__material.opacity_map = None
        elif os.path.isfile(self.inputs["opacity"].file_path):
            self.__material.opacity_map = self.inputs["opacity"].file_path
        self.material_changed.emit()

    @slot
    def slot_opacity_value_changed(self, value):
        self.__material.opacity = value
        self.material_changed.emit()

    @slot
    def slot_opacity_map_changed(self, file_path):
        self.__material.opacity_map = file_path
        self.material_changed.emit()

    @slot
    def slot_height_scale_changed(self, value):
        self.__material.height_scale = value
        self.material_changed.emit()

    @slot
    def slot_fog_changed(self, state):
        self.__material.fog = (state == Qt.CheckState.Checked.value)
        self.material_changed.emit()

    @slot
    def slot_recv_shadows_changed(self, state):
        self.__material.recv_shadows = (state == Qt.CheckState.Checked.value)
        self.material_changed.emit()

    @slot
    def slot_cast_shadows_changed(self, state):
        self.__material.cast_shadows = (state == Qt.CheckState.Checked.value)
        self.material_changed.emit()

    @slot
    def slot_ao_map_checked_changed(self, checked):
        if checked and os.path.isfile(self.inputs["ao_map"].file_path):
            self.__material.ao_map = self.inputs["ao_map"].file_path
        else:
            self.__material.ao_map = None
        self.material_changed.emit()

    @slot
    def slot_ao_map_changed(self, ao_map):
        self.__material.ao_map = ao_map
        self.material_changed.emit()

    @slot
    def slot_normal_map_checked_changed(self, checked):
        if checked and os.path.isfile(self.inputs["normal_map"].file_path):
            self.__material.normal_map = self.inputs["normal_map"].file_path
        else:
            self.__material.normal_map = None
        self.material_changed.emit()

    @slot
    def slot_normal_map_changed(self, normal_map):
        self.__material.normal_map = normal_map
        self.material_changed.emit()

    @slot
    def slot_height_map_checked_changed(self, checked):
        if checked and os.path.isfile(self.inputs["height_map"].file_path):
            self.__material.height_map = self.inputs["height_map"].file_path
        else:
            self.__material.height_map = None
        self.material_changed.emit()

    @slot
    def slot_height_map_changed(self, height_map):
        self.__material.height_map = height_map
        self.material_changed.emit()

    @slot
    def slot_arm_map_checked_changed(self, checked):
        if checked and os.path.isfile(self.inputs["arm_map"].file_path):
            self.__material.arm_map = self.inputs["arm_map"].file_path
        else:
            self.__material.arm_map = None
        self.material_changed.emit()

    @slot
    def slot_arm_map_changed(self, arm_map):
        self.__material.arm_map = arm_map
        self.material_changed.emit()

class MainWindow(QDialog):

    def __init__(self, parent:QWidget=None):
        QDialog.__init__(self, parent=parent)
        self.setWindowFlags(Qt.WindowType.WindowMinMaxButtonsHint | Qt.WindowType.WindowCloseButtonHint)
        self.setWindowTitle('glass_engine 材质编辑概念演示')

        self_folder = os.path.dirname(os.path.abspath(__file__))
        self.setWindowIcon(QIcon(self_folder + "/../images/glass_engine_logo64.png"))

        self.scene, self.camera, dir_light = ModelView()
        self.scene.skydome = "https://dl.polyhaven.org/file/ph-assets/HDRIs/extra/Tonemapped%20JPG/industrial_sunset_puresky.jpg"
        self.camera.screen.bloom = True
        
        sphere = Sphere()
        sphere.material.opacity = 1
        sphere.color = glm.vec4(1,1,1,0)
        self.scene.add(sphere)

        self_path = os.path.dirname(os.path.abspath(__file__))
        model = Model(self_path + "/assets/models/jet/11805_airplane_v2_L2.obj")
        model["root"].scale = 1/300
        model["root"].position.z -= 1
        model.position.y = 5
        model.yaw = 45
        self.scene.add(model)

        floor = Floor()
        floor.position.z = -1
        self.scene.add(floor)

        self.material_editor = MaterialEditor(sphere.material)
        self.splitter = QSplitter()
        self.splitter.addWidget(self.camera.screen)
        self.splitter.addWidget(self.material_editor)
        self.material_editor.material_changed.connect(self.slot_material_changed)

        hlayout = QHBoxLayout()
        hlayout.addWidget(self.splitter)
        self.setLayout(hlayout)

    def slot_material_changed(self):
        self.camera.screen.update()

def demo_material():
    app = QApplication(sys.argv)
    apply_stylesheet(app, theme='dark_amber.xml')

    main_window = MainWindow()
    main_window.show()

    app.exec()